"""
Implementing a basic command-line interface.
"""

## This file is available from https://github.com/adbar/htmldate
## under GNU GPL v3 license

import argparse
import logging
import sys

from platform import python_version
from typing import Any, Optional

from . import __version__
from .core import find_date
from .utils import fetch_url
from .settings import MIN_FILE_SIZE, MAX_FILE_SIZE


def examine(
    htmlstring: Optional[str],
    extensive_bool: bool = True,
    original_date: bool = False,
    verbose_flag: bool = False,
    mindate: Optional[Any] = None,
    maxdate: Optional[Any] = None,
) -> Optional[str]:
    """Generic safeguards and triggers"""
    # safety check
    if htmlstring is None:
        sys.stderr.write("# ERROR: empty document\n")
    elif len(htmlstring) > MAX_FILE_SIZE:
        sys.stderr.write("# ERROR: file too large\n")
    elif len(htmlstring) < MIN_FILE_SIZE:
        sys.stderr.write("# ERROR: file too small\n")
    else:
        return find_date(
            htmlstring,
            extensive_search=extensive_bool,
            original_date=original_date,
            verbose=verbose_flag,
            min_date=mindate,
            max_date=maxdate,
        )
    return None


def parse_args(args: Any) -> Any:
    """Define parser for command-line arguments"""
    argsparser = argparse.ArgumentParser()
    argsparser.add_argument(
        "-f", "--fast", help="fast mode: disable extensive search", action="store_false"
    )
    argsparser.add_argument(
        "-i",
        "--inputfile",
        help="""name of input file for batch processing
                            (similar to wget -i)""",
        type=str,
    )
    argsparser.add_argument(
        "--original", help="original date prioritized", action="store_true"
    )
    argsparser.add_argument(
        "-min", "--mindate", help="earliest acceptable date (YYYY-MM-DD)", type=str
    )
    argsparser.add_argument(
        "-max", "--maxdate", help="latest acceptable date (YYYY-MM-DD)", type=str
    )
    argsparser.add_argument("-u", "--URL", help="custom URL download", type=str)
    argsparser.add_argument(
        "-v", "--verbose", help="increase output verbosity", action="store_true"
    )
    argsparser.add_argument(
        "--version",
        help="show version information and exit",
        action="version",
        version=f"Htmldate {__version__} - Python {python_version()}",
    )
    return argsparser.parse_args()


def process_args(args: Any) -> None:
    """Process the arguments passed on the command-line."""
    # verbosity
    if args.verbose is True:
        logging.basicConfig(level=logging.DEBUG)

    # input type
    if not args.inputfile:
        # URL as input
        if args.URL:
            htmlstring = fetch_url(args.URL)
            if htmlstring is None:
                sys.exit(f"# ERROR no valid result for url: {args.URL}" + "\n")
        # unicode check
        else:
            try:
                htmlstring = sys.stdin.read()
            except UnicodeDecodeError as err:
                sys.exit(f"# ERROR system/buffer encoding: {str(err)}" + "\n")
        result = examine(
            htmlstring,
            extensive_bool=args.fast,
            original_date=args.original,
            verbose_flag=args.verbose,
            maxdate=args.maxdate,
        )
        if result is not None:
            sys.stdout.write(result + "\n")

    # process input file line by line
    else:
        with open(args.inputfile, mode="r", encoding="utf-8") as inputfile:
            for line in inputfile:
                htmltext = fetch_url(line.strip())
                result = examine(
                    htmltext,
                    extensive_bool=args.fast,
                    original_date=args.original,
                    verbose_flag=args.verbose,
                    mindate=args.mindate,
                    maxdate=args.maxdate,
                )
                if result is None:
                    result = "None"
                sys.stdout.write(line.strip() + "\t" + result + "\n")


def main() -> None:
    """Run as a command-line utility."""
    # arguments
    args = parse_args(sys.argv[1:])
    # process input on STDIN
    process_args(args)


if __name__ == "__main__":
    main()



# pylint:disable-msg=E0611,I1101
"""
Custom parsers and XPath expressions for date extraction
"""

## This file is available from https://github.com/adbar/htmldate
## under GNU GPL v3 license

# standard
import logging
import re

from datetime import datetime
from functools import lru_cache
from typing import List, Optional, Pattern, Tuple

# coverage for date parsing
from dateparser import DateDataParser  # type: ignore  # third-party, slow
from dateparser_data.settings import default_parsers

from dateutil.parser import parse as dateutil_parse

from lxml.html import HtmlElement  # type: ignore

try:
    datetime.fromisoformat  # type: ignore[attr-defined]
except AttributeError:  # Python 3.6
    from backports.datetime_fromisoformat import MonkeyPatch  # type: ignore

    MonkeyPatch.patch_fromisoformat()

# own
from .settings import CACHE_SIZE
from .validators import convert_date, date_validator


LOGGER = logging.getLogger(__name__)

EXTERNAL_PARSER = DateDataParser(
    languages=None,
    locales=None,
    region=None,
    settings={
        "NORMALIZE": True,  # False may be faster
        "PARSERS": [
            p
            for p in default_parsers
            if p not in ("no-spaces-time", "relative-time", "timestamp")
        ],
        "PREFER_DATES_FROM": "past",
        "PREFER_LOCALE_DATE_ORDER": True,
        "RETURN_AS_TIMEZONE_AWARE": False,
        "STRICT_PARSING": True,
    },
)


FAST_PREPEND = ".//*[(self::div or self::li or self::p or self::span)]"
# self::b or self::em or self::font or self::i or self::strong
SLOW_PREPEND = ".//*"

DATE_EXPRESSIONS = """
    [contains(translate(@id, "D", "d"), 'date')
    or contains(translate(@class, "D", "d"), 'date')
    or contains(translate(@itemprop, "D", "d"), 'date')
    or contains(translate(@id, "D", "d"), 'datum')
    or contains(translate(@class, "D", "d"), 'datum')
    or contains(@id, 'time') or contains(@class, 'time')
    or @class='meta' or contains(translate(@id, "M", "m"), 'metadata')
    or contains(translate(@class, "M", "m"), 'meta-')
    or contains(translate(@class, "M", "m"), '-meta')
    or contains(translate(@id, "M", "m"), '-meta')
    or contains(translate(@class, "M", "m"), '_meta')
    or contains(translate(@class, "M", "m"), 'postmeta')
    or contains(@class, 'info') or contains(@class, 'post_detail')
    or contains(@class, 'block-content')
    or contains(@class, 'byline') or contains(@class, 'subline')
    or contains(@class, 'posted') or contains(@class, 'submitted')
    or contains(@class, 'created-post')
    or contains(@id, 'publish') or contains(@class, 'publish')
    or contains(@class, 'publication')
    or contains(@class, 'author') or contains(@class, 'autor')
    or contains(@class, 'field-content')
    or contains(@class, 'fa-clock-o') or contains(@class, 'fa-calendar')
    or contains(@class, 'fecha') or contains(@class, 'parution')
    or contains(@class, 'footer') or contains(@id, 'footer')]
    |
    .//footer|.//small
    """
# further tests needed:
# or contains(@class, 'article')
# or contains(@id, 'lastmod') or contains(@class, 'updated')

FREE_TEXT_EXPRESSIONS = FAST_PREPEND + "/text()"
MAX_TEXT_SIZE = 48

# discard parts of the webpage
# archive.org banner inserts
DISCARD_EXPRESSIONS = """.//div[@id="wm-ipp-base" or @id="wm-ipp"]"""
# not discarded for consistency (see above):
# .//footer
# .//*[(self::div or self::section)][@id="footer" or @class="footer"]

# regex cache
YMD_NO_SEP_PATTERN = re.compile(r"\b(\d{8})\b")
YMD_PATTERN = re.compile(
    r"(?:\D|^)(?P<year>\d{4})[\-/.](?P<month>\d{1,2})[\-/.](?P<day>\d{1,2})(?:\D|$)|"
    r"(?:\D|^)(?P<day2>\d{1,2})[\-/.](?P<month2>\d{1,2})[\-/.](?P<year2>\d{2,4})(?:\D|$)"
)
YM_PATTERN = re.compile(
    r"(?:\D|^)(?P<year>\d{4})[\-/.](?P<month>\d{1,2})(?:\D|$)|"
    r"(?:\D|^)(?P<month2>\d{1,2})[\-/.](?P<year2>\d{4})(?:\D|$)"
)

REGEX_MONTHS = """
January?|February?|March|A[pv]ril|Ma[iy]|Jun[ei]|Jul[iy]|August|September|O[ck]tober|November|De[csz]ember|
Jan|Feb|M[aä]r|Apr|Jun|Jul|Aug|Sep|O[ck]t|Nov|De[cz]|
Januari|Februari|Maret|Mei|Agustus|
Jänner|Feber|März|
janvier|février|mars|juin|juillet|aout|septembre|octobre|novembre|décembre|
Ocak|Şubat|Mart|Nisan|Mayıs|Haziran|Temmuz|Ağustos|Eylül|Ekim|Kasım|Aralık|
Oca|Şub|Mar|Nis|Haz|Tem|Ağu|Eyl|Eki|Kas|Ara
"""  # todo: check "août"
LONG_TEXT_PATTERN = re.compile(
    rf"""(?P<month>{REGEX_MONTHS})\s
(?P<day>[0-9]{{1,2}})(?:st|nd|rd|th)?,? (?P<year>[0-9]{{4}})|(?P<day2>[0-9]{{1,2}})(?:st|nd|rd|th|\.)? (?:of )?
(?P<month2>{REGEX_MONTHS})[,.]? (?P<year2>[0-9]{{4}})""".replace(
        "\n", ""
    ),
    re.I,
)

COMPLETE_URL = re.compile(r"\D([0-9]{4})[/_-]([0-9]{1,2})[/_-]([0-9]{1,2})(?:\D|$)")
PARTIAL_URL = re.compile(r"\D([0-9]{4})[/_-]([0-9]{2})(?:\D|$)")

JSON_MODIFIED = re.compile(r'"dateModified": ?"([0-9]{4}-[0-9]{2}-[0-9]{2})', re.I)
JSON_PUBLISHED = re.compile(r'"datePublished": ?"([0-9]{4}-[0-9]{2}-[0-9]{2})', re.I)
TIMESTAMP_PATTERN = re.compile(
    r"([0-9]{4}-[0-9]{2}-[0-9]{2}|[0-9]{2}\.[0-9]{2}\.[0-9]{4}).[0-9]{2}:[0-9]{2}:[0-9]{2}"
)

# English, French, German, Indonesian and Turkish dates cache
TEXT_MONTHS = {
    # January
    "januar": "01",
    "jänner": "01",
    "january": "01",
    "januari": "01",
    "janvier": "01",
    "jan": "01",
    "ocak": "01",
    "oca": "01",
    # February
    "februar": "02",
    "feber": "02",
    "february": "02",
    "februari": "02",
    "février": "02",
    "feb": "02",
    "şubat": "02",
    "şub": "02",
    # March
    "märz": "03",
    "march": "03",
    "maret": "03",
    "mar": "03",
    "mär": "03",
    "mart": "03",
    "mars": "03",
    # April
    "april": "04",
    "apr": "04",
    "avril": "04",
    "nisan": "04",
    "nis": "04",
    # May
    "mai": "05",
    "may": "05",
    "mei": "05",
    "mayıs": "05",
    # June
    "juni": "06",
    "june": "06",
    "juin": "06",
    "jun": "06",
    "haziran": "06",
    "haz": "06",
    # July
    "juli": "07",
    "july": "07",
    "juillet": "07",
    "jul": "07",
    "temmuz": "07",
    "tem": "07",
    # August
    "august": "08",
    "agustus": "08",
    "aug": "08",
    "ağustos": "08",
    "ağu": "08",
    "aout": "08",
    # "août": "08",
    # September
    "september": "09",
    "septembre": "09",
    "sep": "09",
    "eylül": "09",
    "eyl": "09",
    # October
    "oktober": "10",
    "october": "10",
    "octobre": "10",
    "oct": "10",
    "okt": "10",
    "ekim": "10",
    "eki": "10",
    # November
    "november": "11",
    "nov": "11",
    "kasım": "11",
    "kas": "11",
    "novembre": "11",
    # December
    "dezember": "12",
    "december": "12",
    "desember": "12",
    "décembre": "12",
    "dec": "12",
    "dez": "12",
    "aralık": "12",
    "ara": "12",
}

TEXT_DATE_PATTERN = re.compile(r"[.:,_/ -]|^\d+$")
NO_TEXT_DATE_PATTERN = re.compile(
    r"\d{3,}\D+\d{3,}|\d{2}:\d{2}(:| )|\+\d{2}\D+|\D*\d{4}\D*$"
)
# leads to errors: \D+\d{3,}\D+

DISCARD_PATTERNS = re.compile(
    r"[$€¥Ұ£¢₽₱฿#]|CNY|EUR|GBP|JPY|USD|http|\.(com|net|org)|IBAN|\+\d{2}\b"
)
# further testing required:
# \d[,.]\d+  # currency amounts
# \b\d{5}\s  # postal codes

# use of regex module for speed?
TEXT_PATTERNS = re.compile(
    r'(?:date[^0-9"]{,20}|updated|published) *?(?:in)? *?:? *?([0-9]{1,4})[./]([0-9]{1,2})[./]([0-9]{2,4})|'  # EN
    r"(?:Datum|Stand): ?([0-9]{1,2})\.([0-9]{1,2})\.([0-9]{2,4})|"  # DE
    r"(?:güncellen?me|yayı(?:m|n)lan?ma) *?(?:tarihi)? *?:? *?([0-9]{1,2})[./]([0-9]{1,2})[./]([0-9]{2,4})|"
    r"([0-9]{1,2})[./]([0-9]{1,2})[./]([0-9]{2,4}) *?(?:'de|'da|'te|'ta|’de|’da|’te|’ta|tarihinde) *(?:güncellendi|yayı(?:m|n)landı)",  # TR
    re.I,
)

# core patterns
THREE_COMP_REGEX_A = re.compile(r"([0-3]?[0-9])[/.-]([01]?[0-9])[/.-]([0-9]{4})")
THREE_COMP_REGEX_B = re.compile(r"([0-3]?[0-9])[/.-]([01]?[0-9])[/.-]([0-9]{2})")
TWO_COMP_REGEX = re.compile(r"([0-3]?[0-9])[/.-]([0-9]{4})")

# extensive search patterns
YEAR_PATTERN = re.compile(r"^\D?(199[0-9]|20[0-9]{2})")
COPYRIGHT_PATTERN = re.compile(
    r"(?:©|\&copy;|Copyright|\(c\))\D*(?:[12][0-9]{3}-)?([12][0-9]{3})\D"
)
THREE_PATTERN = re.compile(r"/([0-9]{4}/[0-9]{2}/[0-9]{2})[01/]")
THREE_CATCH = re.compile(r"([0-9]{4})/([0-9]{2})/([0-9]{2})")
THREE_LOOSE_PATTERN = re.compile(r"\D([0-9]{4}[/.-][0-9]{2}[/.-][0-9]{2})\D")
THREE_LOOSE_CATCH = re.compile(r"([0-9]{4})[/.-]([0-9]{2})[/.-]([0-9]{2})")
SELECT_YMD_PATTERN = re.compile(r"\D([0-3]?[0-9][/.-][01]?[0-9][/.-][0-9]{4})\D")
SELECT_YMD_YEAR = re.compile(r"(19[0-9]{2}|20[0-9]{2})\D?$")
YMD_YEAR = re.compile(r"^([0-9]{4})")
DATESTRINGS_PATTERN = re.compile(
    r"(\D19[0-9]{2}[01][0-9][0-3][0-9]\D|\D20[0-9]{2}[01][0-9][0-3][0-9]\D)"
)
DATESTRINGS_CATCH = re.compile(r"([12][0-9]{3})([01][0-9])([0-3][0-9])")
SLASHES_PATTERN = re.compile(r"\D([0-3]?[0-9][/.][01]?[0-9][/.][0129][0-9])\D")
SLASHES_YEAR = re.compile(r"([0-9]{2})$")
YYYYMM_PATTERN = re.compile(r"\D([12][0-9]{3}[/.-][01][0-9])\D")
YYYYMM_CATCH = re.compile(r"([12][0-9]{3})[/.-]([01][0-9])")
MMYYYY_PATTERN = re.compile(r"\D([01]?[0-9][/.-][12][0-9]{3})\D")
MMYYYY_YEAR = re.compile(r"([12][0-9]{3})\D?$")
SIMPLE_PATTERN = re.compile(r"\D(199[0-9]|20[0-9]{2})\D")


def discard_unwanted(tree: HtmlElement) -> Tuple[HtmlElement, List[HtmlElement]]:
    """Delete unwanted sections of an HTML document and return them as a list"""
    my_discarded = []
    for subtree in tree.xpath(DISCARD_EXPRESSIONS):
        my_discarded.append(subtree)
        subtree.getparent().remove(subtree)
    return tree, my_discarded


def extract_url_date(
    testurl: str, outputformat: str, min_date: datetime, max_date: datetime
) -> Optional[str]:
    """Extract the date out of an URL string complying with the Y-M-D format"""
    match = COMPLETE_URL.search(testurl)
    if match:
        LOGGER.debug("found date in URL: %s", match[0])
        try:
            dateobject = datetime(int(match[1]), int(match[2]), int(match[3]))
            if (
                date_validator(
                    dateobject, outputformat, earliest=min_date, latest=max_date
                )
                is True
            ):
                return dateobject.strftime(outputformat)
        except ValueError as err:
            LOGGER.debug("conversion error: %s %s", match[0], err)
    return None


def extract_partial_url_date(
    testurl: str, outputformat: str, min_date: datetime, max_date: datetime
) -> Optional[str]:
    """Extract an approximate date out of an URL string in Y-M format"""
    match = PARTIAL_URL.search(testurl)
    if match:
        dateresult = match[0] + "/01"
        LOGGER.debug("found partial date in URL: %s", dateresult)
        try:
            dateobject = datetime(int(match[1]), int(match[2]), 1)
            if (
                date_validator(
                    dateobject, outputformat, earliest=min_date, latest=max_date
                )
                is True
            ):
                return dateobject.strftime(outputformat)
        except ValueError as err:
            LOGGER.debug("conversion error: %s %s", dateresult, err)
    return None


def correct_year(year: int) -> int:
    """Adapt year from YY to YYYY format"""
    if year < 100:
        year += 1900 if year >= 90 else 2000
    return year


def try_swap_values(day: int, month: int) -> Tuple[int, int]:
    """Swap day and month values if it seems feaaible."""
    # If month is more than 12, swap it with the day
    if month > 12 and day <= 12:
        day, month = month, day
    return day, month


def regex_parse(string: str) -> Optional[datetime]:
    """Try full-text parse for date elements using a series of regular expressions
    with particular emphasis on English, French, German and Turkish"""
    # https://github.com/vi3k6i5/flashtext ?
    # multilingual day-month-year + American English patterns
    match = LONG_TEXT_PATTERN.search(string)
    if not match:
        return None
    # process and return
    try:
        if match.lastgroup == "year":
            day, month, year = (
                int(match.group("day")),
                int(TEXT_MONTHS[match.group("month").lower().strip(".")]),
                int(match.group("year")),
            )
        else:
            day, month, year = (
                int(match.group("day2")),
                int(TEXT_MONTHS[match.group("month2").lower().strip(".")]),
                int(match.group("year2")),
            )
        year = correct_year(year)
        day, month = try_swap_values(day, month)
        dateobject = datetime(year, month, day)
    except ValueError:
        return None
    else:
        LOGGER.debug("multilingual text found: %s", dateobject)
        return dateobject


def custom_parse(
    string: str, outputformat: str, min_date: datetime, max_date: datetime
) -> Optional[str]:
    """Try to bypass the slow dateparser"""
    LOGGER.debug("custom parse test: %s", string)

    # 1. shortcut
    if string[:4].isdigit():
        candidate = None
        # a. '201709011234' not covered by dateparser, and regex too slow
        if string[4:8].isdigit():
            try:
                candidate = datetime(
                    int(string[:4]), int(string[4:6]), int(string[6:8])
                )
            except ValueError:
                LOGGER.debug("8-digit error: %s", string[:8])  # return None
        # b. much faster than extensive parsing
        else:
            try:
                candidate = datetime.fromisoformat(string)  # type: ignore[attr-defined]
            except ValueError:
                LOGGER.debug("not an ISO date string: %s", string)
                try:
                    candidate = dateutil_parse(string, fuzzy=False)  # ignoretz=True
                except (OverflowError, TypeError, ValueError):
                    LOGGER.debug("dateutil parsing error: %s", string)
        # c. plausibility test
        if candidate is not None and (
            date_validator(candidate, outputformat, earliest=min_date, latest=max_date)
            is True
        ):
            LOGGER.debug("parsing result: %s", candidate)
            return candidate.strftime(outputformat)

    # 2. Try YYYYMMDD, use regex
    match = YMD_NO_SEP_PATTERN.search(string)
    if match:
        try:
            year, month, day = int(match[1][:4]), int(match[1][4:6]), int(match[1][6:8])
            candidate = datetime(year, month, day)
        except ValueError:
            LOGGER.debug("YYYYMMDD value error: %s", match[0])
        else:
            if (
                date_validator(
                    candidate, "%Y-%m-%d", earliest=min_date, latest=max_date
                )
                is True
            ):
                LOGGER.debug("YYYYMMDD match: %s", candidate)
                return candidate.strftime(outputformat)

    # 3. Try the very common YMD, Y-M-D, and D-M-Y patterns
    match = YMD_PATTERN.search(string)
    if match:
        try:
            # YMD
            if match.lastgroup == "day":
                candidate = datetime(
                    int(match.group("year")),
                    int(match.group("month")),
                    int(match.group("day")),
                )
            # DMY
            else:
                day, month, year = (
                    int(match.group("day2")),
                    int(match.group("month2")),
                    int(match.group("year2")),
                )
                year = correct_year(year)
                day, month = try_swap_values(day, month)
                candidate = datetime(year, month, day)
        except ValueError:
            LOGGER.debug("regex value error: %s", match[0])
        else:
            if (
                date_validator(
                    candidate, "%Y-%m-%d", earliest=min_date, latest=max_date
                )
                is True
            ):
                LOGGER.debug("regex match: %s", candidate)
                return candidate.strftime(outputformat)

    # 4. Try the Y-M and M-Y patterns
    match = YM_PATTERN.search(string)
    if match:
        try:
            if match.lastgroup == "month":
                candidate = datetime(
                    int(match.group("year")), int(match.group("month")), 1
                )
            else:
                candidate = datetime(
                    int(match.group("year2")), int(match.group("month2")), 1
                )
        except ValueError:
            LOGGER.debug("Y-M value error: %s", match[0])
        else:
            if (
                date_validator(
                    candidate, "%Y-%m-%d", earliest=min_date, latest=max_date
                )
                is True
            ):
                LOGGER.debug("Y-M match: %s", candidate)
                return candidate.strftime(outputformat)

    # 5. Try the other regex pattern
    dateobject = regex_parse(string)
    if (
        date_validator(dateobject, outputformat, earliest=min_date, latest=max_date)
        is True
    ):
        try:
            LOGGER.debug("custom parse result: %s", dateobject)
            return dateobject.strftime(outputformat)  # type: ignore
        except ValueError as err:
            LOGGER.error("value error during conversion: %s %s", string, err)

    return None


def external_date_parser(string: str, outputformat: str) -> Optional[str]:
    """Use dateutil parser or dateparser module according to system settings"""
    LOGGER.debug("send to external parser: %s", string)
    try:
        target = EXTERNAL_PARSER.get_date_data(string)["date_obj"]
    # 2 types of errors possible
    except (OverflowError, ValueError) as err:  # pragma: no cover
        target = None
        LOGGER.error("external parser error: %s %s", string, err)
    # issue with data type
    return datetime.strftime(target, outputformat) if target is not None else None


@lru_cache(maxsize=CACHE_SIZE)
def try_date_expr(
    string: str,
    outputformat: str,
    extensive_search: bool,
    min_date: datetime,
    max_date: datetime,
) -> Optional[str]:
    """Use a series of heuristics and rules to parse a potential date expression"""
    # trim
    string = " ".join(string.strip()[:MAX_TEXT_SIZE].split())

    # formal constraint: 4 to 18 digits
    if not string or not 4 <= sum(map(str.isdigit, string)) <= 18:
        return None

    # check if string only contains time/single year or digits and not a date
    if NO_TEXT_DATE_PATTERN.match(string):
        return None

    # try to parse using the faster method
    customresult = custom_parse(string, outputformat, min_date, max_date)
    if customresult is not None:
        return customresult

    # use slow but extensive search
    if extensive_search:
        # additional filters to prevent computational cost
        if not TEXT_DATE_PATTERN.search(string) or DISCARD_PATTERNS.search(string):
            return None
        # send to date parser
        dateparser_result = external_date_parser(string, outputformat)
        if date_validator(
            dateparser_result, outputformat, earliest=min_date, latest=max_date
        ):
            return dateparser_result

    return None


def img_search(
    tree: HtmlElement, outputformat: str, min_date: datetime, max_date: datetime
) -> Optional[str]:
    """Skim through image elements"""
    element = tree.find('.//meta[@property="og:image"][@content]')
    if element is not None:
        result = extract_url_date(
            element.get("content"), outputformat, min_date, max_date
        )
        if result is not None:
            return result
    return None


def json_search(
    tree: HtmlElement,
    outputformat: str,
    original_date: bool,
    min_date: datetime,
    max_date: datetime,
) -> Optional[str]:
    """Look for JSON time patterns in JSON sections of the tree"""
    # determine pattern
    json_pattern = JSON_PUBLISHED if original_date else JSON_MODIFIED
    # look throughout the HTML tree
    for elem in tree.xpath(
        './/script[@type="application/ld+json" or @type="application/settings+json"]'
    ):
        if not elem.text or '"date' not in elem.text:
            continue
        json_match = json_pattern.search(elem.text)
        if json_match and date_validator(
            json_match[1], "%Y-%m-%d", earliest=min_date, latest=max_date
        ):
            LOGGER.debug("JSON time found: %s", json_match[0])
            return convert_date(json_match[1], "%Y-%m-%d", outputformat)
    return None


def timestamp_search(
    htmlstring: str, outputformat: str, min_date: datetime, max_date: datetime
) -> Optional[str]:
    """Look for timestamps throughout the web page"""
    tstamp_match = TIMESTAMP_PATTERN.search(htmlstring)
    if tstamp_match and date_validator(
        tstamp_match[1], "%Y-%m-%d", earliest=min_date, latest=max_date
    ):
        LOGGER.debug("time regex found: %s", tstamp_match[0])
        return convert_date(tstamp_match[1], "%Y-%m-%d", outputformat)
    return None


def extract_idiosyncrasy(
    idiosyncrasy: Pattern[str],
    htmlstring: str,
    outputformat: str,
    min_date: datetime,
    max_date: datetime,
) -> Optional[str]:
    """Look for a precise pattern throughout the web page"""
    candidate = None
    match = idiosyncrasy.search(htmlstring)
    parts = [0, 1, 2, 3] if match and match[3] else []  # because len(None) has no len
    try:
        parts = [0, 4, 5, 6] if match and match[6] else parts
    except IndexError:
        LOGGER.debug("index error in idiosyncrasies: %s", match[0])  # type: ignore[index]
    if match and parts:
        if match[parts[1]] is not None and len(match[parts[1]]) == 4:
            candidate = datetime(
                int(match[parts[1]]), int(match[parts[2]]), int(match[parts[3]])
            )
        elif len(match[parts[3]]) in (2, 4):
            # DD/MM/YY
            day, month = try_swap_values(int(match[parts[1]]), int(match[parts[2]]))
            year = correct_year(int(match[parts[3]]))
            try:
                candidate = datetime(year, month, day)
            except ValueError:
                LOGGER.debug("value error in idiosyncrasies: %s", match[0])
    if (
        date_validator(candidate, "%Y-%m-%d", earliest=min_date, latest=max_date)
        is True
    ):
        LOGGER.debug("idiosyncratic pattern found: %s", match[0])  # type: ignore[index]
        return candidate.strftime(outputformat)  # type: ignore
    return None


def idiosyncrasies_search(
    htmlstring: str, outputformat: str, min_date: datetime, max_date: datetime
) -> Optional[str]:
    """Look for author-written dates throughout the web page"""
    result = None
    # EN+DE+TR
    result = extract_idiosyncrasy(
        TEXT_PATTERNS, htmlstring, outputformat, min_date, max_date
    )
    return result



"""
Htmldate extracts original and updated publication dates from URLs and web pages.
"""

# meta
__title__ = "htmldate"
__author__ = "Adrien Barbaresi"
__license__ = "GNU GPL v3"
__copyright__ = "Copyright 2017-2023, Adrien Barbaresi"
__version__ = "1.4.1"


import logging

from .core import find_date

logging.getLogger(__name__).addHandler(logging.NullHandler())



# pylint:disable-msg=E0611,I1101
"""Module bundling all functions needed to determine the date of HTML strings
or LXML trees.
"""

## This file is available from https://github.com/adbar/htmldate
## under GNU GPL v3 license

# standard
import logging
import re

from collections import Counter
from copy import deepcopy
from datetime import datetime
from functools import lru_cache, partial
from typing import Match, Optional, Pattern, Tuple, Counter as Counter_Type

from lxml.html import HtmlElement, tostring  # type: ignore

# own
from .extractors import (
    discard_unwanted,
    extract_url_date,
    extract_partial_url_date,
    idiosyncrasies_search,
    img_search,
    json_search,
    regex_parse,
    timestamp_search,
    try_date_expr,
    DATE_EXPRESSIONS,
    FAST_PREPEND,
    SLOW_PREPEND,
    FREE_TEXT_EXPRESSIONS,
    MAX_TEXT_SIZE,
    YEAR_PATTERN,
    YMD_PATTERN,
    COPYRIGHT_PATTERN,
    THREE_PATTERN,
    THREE_CATCH,
    THREE_LOOSE_PATTERN,
    THREE_LOOSE_CATCH,
    SELECT_YMD_PATTERN,
    SELECT_YMD_YEAR,
    YMD_YEAR,
    DATESTRINGS_PATTERN,
    DATESTRINGS_CATCH,
    SLASHES_PATTERN,
    SLASHES_YEAR,
    YYYYMM_PATTERN,
    YYYYMM_CATCH,
    MMYYYY_PATTERN,
    MMYYYY_YEAR,
    SIMPLE_PATTERN,
    THREE_COMP_REGEX_A,
    THREE_COMP_REGEX_B,
    TWO_COMP_REGEX,
)
from .settings import CACHE_SIZE, HTML_CLEANER, MAX_POSSIBLE_CANDIDATES
from .utils import load_html
from .validators import (
    check_extracted_reference,
    compare_values,
    convert_date,
    date_validator,
    filter_ymd_candidate,
    get_min_date,
    get_max_date,
    output_format_validator,
    plausible_year_filter,
)


LOGGER = logging.getLogger(__name__)


def logstring(element: HtmlElement) -> str:
    """Format the element to be logged to a string."""
    return tostring(element, pretty_print=False, encoding="unicode").strip()  # type: ignore


DATE_ATTRIBUTES = {
    "analyticsattributes.articledate",
    "article.created",
    "article_date_original",
    "article:post_date",
    "article.published",
    "article:published",
    "article:published_time",
    "article:publicationdate",
    "bt:pubdate",
    "citation_date",
    "citation_publication_date",
    "content_create_date",
    "created",
    "cxenseparse:recs:publishtime",
    "date",
    "date_created",
    "date_published",
    "datecreated",
    "dateposted",
    "datepublished",
    # Dublin Core: https://wiki.whatwg.org/wiki/MetaExtensions
    "dc.date",
    "dc.created",
    "dc.date.created",
    "dc.date.issued",
    "dc.date.publication",
    "dcsext.articlefirstpublished",
    "dcterms.created",
    "dcterms.date",
    "dcterms.issued",
    "dc:created",
    "dc:date",
    "displaydate",
    "doc_date",
    "field-name-post-date",
    "gentime",
    "mediator_published_time",
    "meta",  # too loose?
    # Open Graph: https://opengraphprotocol.org/
    "og:published_time",
    "og:article:published_time",
    "originalpublicationdate",
    "parsely-pub-date",
    "pdate",
    "ptime",
    "pubdate",
    "publishdate",
    "publish_date",
    "publish-date",
    "published-date",
    "publication_date",
    "rbpubdate",
    "release_date",
    "rnews:datepublished",
    "sailthru.date",
    "shareaholic:article_published_time",
    "timestamp",
    "twt-published-at",
    "video:release_date",
    "vr:published_time",
}


NAME_MODIFIED = {"lastmod", "lastmodified", "last-modified", "modified", "utime"}


PROPERTY_MODIFIED = {
    "article:modified",
    "article:modified_time",
    "datemodified",
    "dc.modified",
    "dcterms.modified",
    "lastmodified",
    "modified_time",
    "og:article:modified_time",
    "og:updated_time",
    "release_date",
    "revision_date",
    "updated_time",
}


ITEMPROP_ATTRS_ORIGINAL = {"datecreated", "datepublished", "pubyear"}
ITEMPROP_ATTRS_MODIFIED = {"datemodified", "dateupdate"}
ITEMPROP_ATTRS = ITEMPROP_ATTRS_ORIGINAL.union(ITEMPROP_ATTRS_MODIFIED)
CLASS_ATTRS = {"date-published", "published", "time published"}

NON_DIGITS_REGEX = re.compile(r"\D+$")


def examine_date_elements(
    tree: HtmlElement,
    expression: str,
    outputformat: str,
    extensive_search: bool,
    min_date: datetime,
    max_date: datetime,
) -> Optional[str]:
    """Check HTML elements one by one for date expressions"""
    elements = tree.xpath(expression)
    if not elements or len(elements) > MAX_POSSIBLE_CANDIDATES:
        return None
    # loop through the elements to analyze
    attempt = None
    for elem in elements:
        # trim
        text = elem.text_content().strip()
        # simple length heuristic
        if len(text) > 6:  # could be 8 or 9
            # shorten and try the beginning of the string
            # trim non-digits at the end of the string
            text = NON_DIGITS_REGEX.sub("", text[:MAX_TEXT_SIZE])
            LOGGER.debug(
                "analyzing (HTML): %s",
                " ".join(logstring(elem).split())[:100],
            )
            attempt = try_date_expr(
                text, outputformat, extensive_search, min_date, max_date
            )
            if attempt is not None:
                break
        # try link title (Blogspot)
        title_attr = elem.get("title", "").strip()
        if title_attr is not None and len(title_attr) > 0:
            title_attr = NON_DIGITS_REGEX.sub("", title_attr[:MAX_TEXT_SIZE])
            attempt = try_date_expr(
                title_attr, outputformat, extensive_search, min_date, max_date
            )
            if attempt is not None:
                break
    # catchall
    return attempt


def examine_header(
    tree: HtmlElement,
    outputformat: str,
    extensive_search: bool,
    original_date: bool,
    min_date: datetime,
    max_date: datetime,
) -> Optional[str]:
    """
    Parse header elements to find date cues

    :param tree:
        LXML parsed tree object
    :type htmlstring: LXML tree
    :param outputformat:
        Provide a valid datetime format for the returned string
        (see datetime.strftime())
    :type outputformat: string
    :param extensive_search:
        Activate pattern-based opportunistic text search
    :type extensive_search: boolean
    :param original_date:
        Look for original date (e.g. publication date) instead of most recent
        one (e.g. last modified, updated time)
    :type original_date: boolean
    :param min_date:
        Set the earliest acceptable date manually (YYYY-MM-DD format)
    :type min_date: string
    :param max_date:
        Set the latest acceptable date manually (YYYY-MM-DD format)
    :type max_date: string
    :return: Returns a valid date expression as a string, or None

    """
    headerdate, reserve = None, None
    tryfunc = partial(
        try_date_expr,
        outputformat=outputformat,
        extensive_search=extensive_search,
        min_date=min_date,
        max_date=max_date,
    )
    # loop through all meta elements
    for elem in tree.iterfind(".//meta"):
        # safeguard
        if not elem.attrib or (
            not "content" in elem.attrib and not "datetime" in elem.attrib
        ):
            continue
        # name attribute, most frequent
        if "name" in elem.attrib:
            attribute = elem.get("name").lower()
            # url
            if attribute == "og:url":
                reserve = extract_url_date(
                    elem.get("content"), outputformat, min_date, max_date
                )
            # date
            elif attribute in DATE_ATTRIBUTES:
                LOGGER.debug("examining meta name: %s", logstring(elem))
                headerdate = tryfunc(elem.get("content"))
            # modified
            elif attribute in NAME_MODIFIED:
                LOGGER.debug("examining meta name: %s", logstring(elem))
                if not original_date:
                    headerdate = tryfunc(elem.get("content"))
                else:
                    reserve = tryfunc(elem.get("content"))
        # property attribute
        elif "property" in elem.attrib:
            attribute = elem.get("property").lower()
            # original date or modified date: override published_time
            if (original_date and attribute in DATE_ATTRIBUTES) or (
                not original_date
                and (attribute in PROPERTY_MODIFIED or attribute in DATE_ATTRIBUTES)
            ):
                LOGGER.debug("examining meta property: %s", logstring(elem))
                headerdate = tryfunc(elem.get("content"))
        # itemprop
        elif "itemprop" in elem.attrib:
            attribute = elem.get("itemprop").lower()
            # original: store / updated: override date
            if attribute in ITEMPROP_ATTRS:
                LOGGER.debug("examining meta itemprop: %s", logstring(elem))
                attempt = tryfunc(elem.get("datetime") or elem.get("content"))
                # store value
                if attempt is not None:
                    if (attribute in ITEMPROP_ATTRS_ORIGINAL and original_date) or (
                        attribute in ITEMPROP_ATTRS_MODIFIED and not original_date
                    ):
                        headerdate = attempt
                    # put on hold: hurts precision
                    # else:
                    #    reserve = attempt
            # reserve with copyrightyear
            elif attribute == "copyrightyear":
                LOGGER.debug("examining meta itemprop: %s", logstring(elem))
                if "content" in elem.attrib:
                    attempt = "-".join([elem.get("content"), "01", "01"])
                    if (
                        date_validator(
                            attempt, "%Y-%m-%d", earliest=min_date, latest=max_date
                        )
                        is True
                    ):
                        reserve = attempt
        # pubdate, relatively rare
        elif "pubdate" in elem.attrib:
            if elem.get("pubdate").lower() == "pubdate":
                LOGGER.debug("examining meta pubdate: %s", logstring(elem))
                headerdate = tryfunc(elem.get("content"))
        # http-equiv, rare
        elif "http-equiv" in elem.attrib:
            attribute = elem.get("http-equiv").lower()
            if attribute == "date":
                LOGGER.debug("examining meta http-equiv: %s", logstring(elem))
                if original_date:
                    headerdate = tryfunc(elem.get("content"))
                else:
                    reserve = tryfunc(elem.get("content"))
            elif attribute == "last-modified":
                LOGGER.debug("examining meta http-equiv: %s", logstring(elem))
                if not original_date:
                    headerdate = tryfunc(elem.get("content"))
                else:
                    reserve = tryfunc(elem.get("content"))
        # exit loop
        if headerdate is not None:
            break
    # if nothing was found, look for lower granularity (so far: "copyright year")
    if headerdate is None and reserve is not None:
        LOGGER.debug("opting for reserve date with less granularity")
        headerdate = reserve
    # return value
    return headerdate


def select_candidate(
    occurrences: Counter_Type[str],
    catch: Pattern[str],
    yearpat: Pattern[str],
    original_date: bool,
    min_date: datetime,
    max_date: datetime,
) -> Optional[Match[str]]:
    """Select a candidate among the most frequent matches"""
    match, year1, year2 = None, None, None
    # LOGGER.debug('occurrences: %s', occurrences)
    if len(occurrences) == 0 or len(occurrences) > MAX_POSSIBLE_CANDIDATES:
        return None
    if len(occurrences) == 1:
        match = catch.search(list(occurrences.keys())[0])
        if match:
            return match
    # select among most frequent
    firstselect = occurrences.most_common(10)
    LOGGER.debug("firstselect: %s", firstselect)
    # sort and find probable candidates
    if original_date:
        bestones = sorted(firstselect)[:2]
    else:
        bestones = sorted(firstselect, reverse=True)[:2]

    first_pattern, first_count = bestones[0][0], bestones[0][1]
    second_pattern, second_count = bestones[1][0], bestones[1][1]
    LOGGER.debug("bestones: %s", bestones)
    # plausibility heuristics
    validation1, validation2 = False, False
    match1 = yearpat.search(first_pattern)
    if match1 is not None:
        year1 = match1[1]
        validation1 = date_validator(year1, "%Y", earliest=min_date, latest=max_date)
    match2 = yearpat.search(second_pattern)
    if match2 is not None:
        year2 = match2[1]
        validation2 = date_validator(year2, "%Y", earliest=min_date, latest=max_date)
    # safety net: plausibility
    if validation1 is True and validation2 is True:
        # same number of occurrences: always take top of the pile?
        if first_count == second_count:
            match = catch.search(first_pattern)
        # safety net: newer date but up to 50% less frequent
        elif year2 != year1 and second_count / first_count > 0.5:
            match = catch.search(second_pattern)
        # not newer or hopefully not significant
        else:
            match = catch.search(first_pattern)
    elif validation1 is False and validation2 is True:
        match = catch.search(second_pattern)
    elif validation1 is True and validation2 is False:
        match = catch.search(first_pattern)
    else:
        LOGGER.debug("no suitable candidate: %s %s", year1, year2)
        return None
    return match


def search_pattern(
    htmlstring: str,
    pattern: Pattern[str],
    catch: Pattern[str],
    yearpat: Pattern[str],
    original_date: bool,
    min_date: datetime,
    max_date: datetime,
) -> Optional[Match[str]]:
    """Chained candidate filtering and selection"""
    candidates = plausible_year_filter(
        htmlstring, pattern=pattern, yearpat=yearpat, earliest=min_date, latest=max_date
    )
    return select_candidate(
        candidates, catch, yearpat, original_date, min_date, max_date
    )


@lru_cache(maxsize=CACHE_SIZE)
def compare_reference(
    reference: int,
    expression: str,
    outputformat: str,
    extensive_search: bool,
    original_date: bool,
    min_date: datetime,
    max_date: datetime,
) -> int:
    """Compare candidate to current date reference (includes date validation and older/newer test)"""
    attempt = try_date_expr(
        expression, outputformat, extensive_search, min_date, max_date
    )
    if attempt is not None:
        return compare_values(reference, attempt, outputformat, original_date)
    return reference


def examine_abbr_elements(
    tree: HtmlElement,
    outputformat: str,
    extensive_search: bool,
    original_date: bool,
    min_date: datetime,
    max_date: datetime,
) -> Optional[str]:
    """Scan the page for abbr elements and check if their content contains an eligible date"""
    elements = tree.findall(".//abbr")
    if elements is not None and len(elements) < MAX_POSSIBLE_CANDIDATES:
        reference = 0
        for elem in elements:
            # data-utime (mostly Facebook)
            if "data-utime" in elem.attrib:
                try:
                    candidate = int(elem.get("data-utime"))
                except ValueError:
                    continue
                LOGGER.debug("data-utime found: %s", candidate)
                # look for original date
                if original_date and (reference == 0 or candidate < reference):
                    reference = candidate
                # look for newest (i.e. largest time delta)
                elif not original_date and candidate > reference:
                    reference = candidate
            # class
            elif elem.get("class") in CLASS_ATTRS:
                # other attributes
                if "title" in elem.attrib:
                    trytext = elem.get("title")
                    LOGGER.debug("abbr published-title found: %s", trytext)
                    # shortcut
                    if original_date:
                        attempt = try_date_expr(
                            trytext, outputformat, extensive_search, min_date, max_date
                        )
                        if attempt is not None:
                            return attempt
                    else:
                        reference = compare_reference(
                            reference,
                            trytext,
                            outputformat,
                            extensive_search,
                            original_date,
                            min_date,
                            max_date,
                        )
                        # faster execution
                        if reference > 0:
                            break
                # dates, not times of the day
                elif elem.text and len(elem.text) > 10:
                    LOGGER.debug("abbr published found: %s", elem.text)
                    reference = compare_reference(
                        reference,
                        elem.text,
                        outputformat,
                        extensive_search,
                        original_date,
                        min_date,
                        max_date,
                    )
        # convert and return
        converted = check_extracted_reference(
            reference, outputformat, min_date, max_date
        )
        if converted is not None:
            return converted
        # try rescue in abbr content
        dateresult = examine_date_elements(
            tree, ".//abbr", outputformat, extensive_search, min_date, max_date
        )
        if dateresult is not None:
            return dateresult
    return None


def examine_time_elements(
    tree: HtmlElement,
    outputformat: str,
    extensive_search: bool,
    original_date: bool,
    min_date: datetime,
    max_date: datetime,
) -> Optional[str]:
    """Scan the page for time elements and check if their content contains an eligible date"""
    elements = tree.findall(".//time")
    if elements is not None and len(elements) < MAX_POSSIBLE_CANDIDATES:
        # scan all the tags and look for the newest one
        reference = 0
        for elem in elements:
            shortcut_flag = False
            # go for datetime
            if "datetime" in elem.attrib and len(elem.get("datetime")) > 6:
                # shortcut: time pubdate
                if "pubdate" in elem.attrib and elem.get("pubdate") == "pubdate":
                    if original_date:
                        shortcut_flag = True
                    LOGGER.debug("time pubdate found: %s", elem.get("datetime"))
                # first choice: entry-date + datetime attribute
                elif "class" in elem.attrib:
                    if elem.get("class").startswith("entry-date") or elem.get(
                        "class"
                    ).startswith("entry-time"):
                        # shortcut
                        if original_date:
                            shortcut_flag = True
                        LOGGER.debug("time/datetime found: %s", elem.get("datetime"))
                    # updated time
                    elif elem.get("class") == "updated" and not original_date:
                        LOGGER.debug(
                            "updated time/datetime found: %s", elem.get("datetime")
                        )
                # datetime attribute
                else:
                    LOGGER.debug("time/datetime found: %s", elem.get("datetime"))
                # analyze attribute
                if shortcut_flag:
                    attempt = try_date_expr(
                        elem.get("datetime"),
                        outputformat,
                        extensive_search,
                        min_date,
                        max_date,
                    )
                    if attempt is not None:
                        return attempt
                else:
                    reference = compare_reference(
                        reference,
                        elem.get("datetime"),
                        outputformat,
                        extensive_search,
                        original_date,
                        min_date,
                        max_date,
                    )
                    if reference > 0:
                        break
            # bare text in element
            elif elem.text is not None and len(elem.text) > 6:
                LOGGER.debug("time/datetime found: %s", elem.text)
                reference = compare_reference(
                    reference,
                    elem.text,
                    outputformat,
                    extensive_search,
                    original_date,
                    min_date,
                    max_date,
                )
            # else...?
        # return
        converted = check_extracted_reference(
            reference, outputformat, min_date, max_date
        )
        if converted is not None:
            return converted
    return None


def normalize_match(match: Optional[Match[str]]) -> Tuple[str, str]:
    """Normalize string output by adding "0" if necessary."""
    day = match[1]  # type: ignore[index]
    if len(day) == 1:
        day = "0" + day
    month = match[2]  # type: ignore[index]
    if len(month) == 1:
        month = "0" + month
    return day, month


def search_page(
    htmlstring: str,
    outputformat: str,
    original_date: bool,
    min_date: datetime,
    max_date: datetime,
) -> Optional[str]:
    """
    Opportunistically search the HTML text for common text patterns

    :param htmlstring:
        The HTML document in string format, potentially cleaned and stripped to
        the core (much faster)
    :type htmlstring: string
    :param outputformat:
        Provide a valid datetime format for the returned string
        (see datetime.strftime())
    :type outputformat: string
    :param original_date:
        Look for original date (e.g. publication date) instead of most recent
        one (e.g. last modified, updated time)
    :type original_date: boolean
    :return: Returns a valid date expression as a string, or None

    """

    # copyright symbol
    LOGGER.debug("looking for copyright/footer information")
    copyear = 0
    bestmatch = search_pattern(
        htmlstring,
        COPYRIGHT_PATTERN,
        YEAR_PATTERN,
        YEAR_PATTERN,
        original_date,
        min_date,
        max_date,
    )
    if bestmatch is not None:
        LOGGER.debug("Copyright detected: %s", bestmatch[0])
        if (
            date_validator(bestmatch[0], "%Y", earliest=min_date, latest=max_date)
            is True
        ):
            LOGGER.debug("copyright year/footer pattern found: %s", bestmatch[0])
            copyear = int(bestmatch[0])

    # 3 components
    LOGGER.debug("3 components")
    # target URL characteristics
    bestmatch = search_pattern(
        htmlstring,
        THREE_PATTERN,
        THREE_CATCH,
        YEAR_PATTERN,
        original_date,
        min_date,
        max_date,
    )
    result = filter_ymd_candidate(
        bestmatch,
        THREE_PATTERN,
        original_date,
        copyear,
        outputformat,
        min_date,
        max_date,
    )
    if result is not None:
        return result

    # more loosely structured data
    bestmatch = search_pattern(
        htmlstring,
        THREE_LOOSE_PATTERN,
        THREE_LOOSE_CATCH,
        YEAR_PATTERN,
        original_date,
        min_date,
        max_date,
    )
    result = filter_ymd_candidate(
        bestmatch,
        THREE_LOOSE_PATTERN,
        original_date,
        copyear,
        outputformat,
        min_date,
        max_date,
    )
    if result is not None:
        return result

    # YYYY-MM-DD/DD-MM-YYYY
    candidates = plausible_year_filter(
        htmlstring,
        pattern=SELECT_YMD_PATTERN,
        yearpat=SELECT_YMD_YEAR,
        earliest=min_date,
        latest=max_date,
    )
    # revert DD-MM-YYYY patterns before sorting
    replacement = {}
    for item in candidates:
        match = THREE_COMP_REGEX_A.match(item)
        day, month = normalize_match(match)
        candidate = "-".join([match[3], month, day])  # type: ignore[index]
        replacement[candidate] = candidates[item]
    candidates = Counter(replacement)
    # select
    bestmatch = select_candidate(
        candidates, YMD_PATTERN, YMD_YEAR, original_date, min_date, max_date
    )
    result = filter_ymd_candidate(
        bestmatch,
        SELECT_YMD_PATTERN,
        original_date,
        copyear,
        outputformat,
        min_date,
        max_date,
    )
    if result is not None:
        return result

    # valid dates strings
    bestmatch = search_pattern(
        htmlstring,
        DATESTRINGS_PATTERN,
        DATESTRINGS_CATCH,
        YEAR_PATTERN,
        original_date,
        min_date,
        max_date,
    )
    result = filter_ymd_candidate(
        bestmatch,
        DATESTRINGS_PATTERN,
        original_date,
        copyear,
        outputformat,
        min_date,
        max_date,
    )
    if result is not None:
        return result

    # DD?/MM?/YY
    candidates = plausible_year_filter(
        htmlstring,
        pattern=SLASHES_PATTERN,
        yearpat=SLASHES_YEAR,
        earliest=min_date,
        latest=max_date,
        incomplete=True,
    )
    # revert DD-MM-YYYY patterns before sorting
    replacement = {}
    for item in candidates:
        match = THREE_COMP_REGEX_B.match(item)
        day, month = normalize_match(match)
        if match[3][0] == "9":  # type: ignore[index]
            year = "19" + match[3]  # type: ignore[index]
        else:
            year = "20" + match[3]  # type: ignore[index]
        candidate = "-".join([year, month, day])
        replacement[candidate] = candidates[item]
    candidates = Counter(replacement)
    bestmatch = select_candidate(
        candidates, YMD_PATTERN, YMD_YEAR, original_date, min_date, max_date
    )
    result = filter_ymd_candidate(
        bestmatch,
        SLASHES_PATTERN,
        original_date,
        copyear,
        outputformat,
        min_date,
        max_date,
    )
    if result is not None:
        return result

    # 2 components
    LOGGER.debug("switching to two components")
    # first option
    bestmatch = search_pattern(
        htmlstring,
        YYYYMM_PATTERN,
        YYYYMM_CATCH,
        YEAR_PATTERN,
        original_date,
        min_date,
        max_date,
    )
    if bestmatch is not None:
        pagedate = "-".join([bestmatch[1], bestmatch[2], "01"])
        if date_validator(
            pagedate, "%Y-%m-%d", earliest=min_date, latest=max_date
        ) is True and (copyear == 0 or int(bestmatch[1]) >= copyear):
            LOGGER.debug('date found for pattern "%s": %s', YYYYMM_PATTERN, pagedate)
            return convert_date(pagedate, "%Y-%m-%d", outputformat)

    # 2 components, second option
    candidates = plausible_year_filter(
        htmlstring,
        pattern=MMYYYY_PATTERN,
        yearpat=MMYYYY_YEAR,
        earliest=min_date,
        latest=max_date,
        incomplete=original_date,
    )
    # revert DD-MM-YYYY patterns before sorting
    replacement = {}
    for item in candidates:
        match = TWO_COMP_REGEX.match(item)
        month = match[1]  # type: ignore[index]
        if len(month) == 1:
            month = "0" + month
        candidate = "-".join([match[2], month, "01"])  # type: ignore[index]
        replacement[candidate] = candidates[item]
    candidates = Counter(replacement)
    # select
    bestmatch = select_candidate(
        candidates, YMD_PATTERN, YMD_YEAR, original_date, min_date, max_date
    )
    result = filter_ymd_candidate(
        bestmatch,
        MMYYYY_PATTERN,
        original_date,
        copyear,
        outputformat,
        min_date,
        max_date,
    )
    if result is not None:
        return result

    # try full-blown text regex on all HTML?
    dateobject = regex_parse(htmlstring)
    # todo: find all candidates and disambiguate?
    if date_validator(
        dateobject, outputformat, earliest=min_date, latest=max_date
    ) is True and (
        copyear == 0 or dateobject.year >= copyear  # type: ignore[union-attr]
    ):
        try:
            LOGGER.debug("regex result on HTML: %s", dateobject)
            return dateobject.strftime(outputformat)  # type: ignore
        except ValueError as err:
            LOGGER.error("value error during conversion: %s %s", dateobject, err)

    # catchall: copyright mention
    if copyear != 0:
        LOGGER.debug("using copyright year as default")
        return convert_date(
            "-".join([str(copyear), "01", "01"]), "%Y-%m-%d", outputformat
        )

    # last resort: 1 component
    LOGGER.debug("switching to one component")
    bestmatch = search_pattern(
        htmlstring,
        SIMPLE_PATTERN,
        YEAR_PATTERN,
        YEAR_PATTERN,
        original_date,
        min_date,
        max_date,
    )
    if bestmatch is not None:
        pagedate = "-".join([bestmatch[0], "01", "01"])
        if (
            date_validator(pagedate, "%Y-%m-%d", earliest=min_date, latest=max_date)
            is True
            and int(bestmatch[0]) >= copyear
        ):
            LOGGER.debug('date found for pattern "%s": %s', SIMPLE_PATTERN, pagedate)
            return convert_date(pagedate, "%Y-%m-%d", outputformat)

    return None


def find_date(
    htmlobject: HtmlElement,
    extensive_search: bool = True,
    original_date: bool = False,
    outputformat: str = "%Y-%m-%d",
    url: Optional[str] = None,
    verbose: bool = False,
    min_date: Optional[datetime] = None,
    max_date: Optional[datetime] = None,
    deferred_url_extractor: bool = False,
) -> Optional[str]:
    """
    Extract dates from HTML documents using markup analysis and text patterns

    :param htmlobject:
        Two possibilities: 1. HTML document (e.g. body of HTTP request or .html-file) in text string
        form or LXML parsed tree or 2. URL string (gets detected automatically)
    :type htmlobject: string or lxml tree
    :param extensive_search:
        Activate pattern-based opportunistic text search
    :type extensive_search: boolean
    :param original_date:
        Look for original date (e.g. publication date) instead of most recent
        one (e.g. last modified, updated time)
    :type original_date: boolean
    :param outputformat:
        Provide a valid datetime format for the returned string
        (see datetime.strftime())
    :type outputformat: string
    :param url:
        Provide an URL manually for pattern-searching in URL
        (in some cases much faster)
    :type url: string
    :param verbose:
        Set verbosity level for debugging
    :type verbose: boolean
    :param min_date:
        Set the earliest acceptable date manually (YYYY-MM-DD format)
    :type min_date: string
    :param max_date:
        Set the latest acceptable date manually (YYYY-MM-DD format)
    :type max_date: string
    :param deferred_url_extractor:
        Use url extractor as backup only to prioritize full expressions,
        e.g. of the type `%Y-%m-%d %H:%M:%S`
    :type deferred_url_extractor: boolean
    :return: Returns a valid date expression as a string, or None
    """

    # init
    if verbose:
        logging.basicConfig(level=logging.DEBUG)
    tree = load_html(htmlobject)

    # safeguards
    if tree is None:
        return None
    if outputformat != "%Y-%m-%d" and not output_format_validator(outputformat):
        return None

    # define time boundaries
    min_date, max_date = get_min_date(min_date), get_max_date(max_date)
    # unclear what this line is for and it impedes type checking:
    # find_date.extensive_search = extensive_search

    # URL
    if url is None:
        # probe for canonical links
        for elem in tree.iterfind('.//link[@rel="canonical"]'):
            url = elem.get("href")
            if url is not None:
                # break out of the loop
                break

    # direct processing of URL info
    if not deferred_url_extractor and url is not None:
        dateresult = extract_url_date(url, outputformat, min_date, max_date)
        if dateresult is not None:
            return dateresult

    # first, try header
    header_result = examine_header(
        tree, outputformat, extensive_search, original_date, min_date, max_date
    )
    if header_result is not None:
        return header_result

    # try to use JSON data
    json_result = json_search(tree, outputformat, original_date, min_date, max_date)
    if json_result is not None:
        return json_result

    # URL
    if url is None:
        # probe for canonical links
        for elem in tree.iterfind('.//link[@rel="canonical"]'):
            url = elem.get("href")
            if url is not None:
                # break out of the loop
                break

    # deferred processing of URL info (may be moved even further down if necessary)
    if url is not None and deferred_url_extractor:
        dateresult = extract_url_date(url, outputformat, min_date, max_date)
        if dateresult is not None:
            return dateresult

    # try abbr elements
    abbr_result = examine_abbr_elements(
        tree, outputformat, extensive_search, original_date, min_date, max_date
    )
    if abbr_result is not None:
        return abbr_result

    # expressions + text_content
    if extensive_search:
        date_expr = SLOW_PREPEND + DATE_EXPRESSIONS
    else:
        date_expr = FAST_PREPEND + DATE_EXPRESSIONS

    # first try in pruned tree
    search_tree, discarded = discard_unwanted(deepcopy(tree))
    dateresult = examine_date_elements(
        search_tree, date_expr, outputformat, extensive_search, min_date, max_date
    )
    if dateresult is not None:
        return dateresult

    # TODO: decide on this
    # search in discarded parts (e.g. archive.org-banner)
    # for subtree in discarded:
    #    dateresult = examine_date_elements(subtree, DATE_EXPRESSIONS,
    #        outputformat, extensive_search, min_date, max_date)
    #    if dateresult is not None:
    #        return dateresult

    # try time elements
    time_result = examine_time_elements(
        search_tree, outputformat, extensive_search, original_date, min_date, max_date
    )
    if time_result is not None:
        return time_result

    # clean before string search
    try:
        cleaned_html = HTML_CLEANER.clean_html(tree)
    # rare LXML error: no NULL bytes or control characters
    except ValueError:  # pragma: no cover
        cleaned_html = tree
        LOGGER.error("lxml cleaner error")

    # robust conversion to string
    try:
        htmlstring = tostring(cleaned_html, pretty_print=False, encoding="unicode")
    except UnicodeDecodeError:
        htmlstring = tostring(cleaned_html, pretty_print=False).decode(
            "utf-8", "ignore"
        )
    # remove comments by hand as faulty in lxml?
    # htmlstring = re.sub(r'<!--.+?-->', '', htmlstring, flags=re.DOTALL)

    # date regex timestamp rescue
    timestamp_result = timestamp_search(htmlstring, outputformat, min_date, max_date)
    if timestamp_result is not None:
        return timestamp_result

    # precise patterns and idiosyncrasies
    text_result = idiosyncrasies_search(htmlstring, outputformat, min_date, max_date)
    if text_result is not None:
        return text_result

    # title
    for title_elem in tree.iter("title", "h1"):
        attempt = try_date_expr(
            title_elem.text_content(),
            outputformat,
            extensive_search,
            min_date,
            max_date,
        )
        if attempt is not None:
            return attempt

    # last try: URL 2
    if url is not None:
        dateresult = extract_partial_url_date(url, outputformat, min_date, max_date)
        if dateresult is not None:
            return dateresult

    # try image elements
    img_result = img_search(tree, outputformat, min_date, max_date)
    if img_result is not None:
        return img_result

    # last resort
    if extensive_search:
        LOGGER.debug("extensive search started")
        # TODO: further tests & decide according to original_date
        reference = 0
        for segment in cleaned_html.xpath(FREE_TEXT_EXPRESSIONS):
            segment = segment.strip()
            # basic filter: minimum could be 8 or 9
            if not 6 < len(segment) < MAX_TEXT_SIZE:
                continue
            reference = compare_reference(
                reference,
                segment,
                outputformat,
                extensive_search,
                original_date,
                min_date,
                max_date,
            )
        # return
        converted = check_extracted_reference(
            reference, outputformat, min_date, max_date
        )
        if converted is not None:
            return converted
        # search page HTML
        return search_page(htmlstring, outputformat, original_date, min_date, max_date)

    return None



# pylint:disable-msg=E0611
"""
Listing a series of settings that are applied module-wide.
"""

## This file is available from https://github.com/adbar/htmldate
## under GNU GPL v3 license

from datetime import datetime

from lxml.html.clean import Cleaner  # type: ignore


# Function cache
CACHE_SIZE: int = 8192

# Download
MAX_FILE_SIZE: int = 20000000
MIN_FILE_SIZE: int = 10

# Plausible dates
# earliest possible date to take into account (inclusive)
MIN_DATE: datetime = datetime(1995, 1, 1)

# set an upper limit to the number of candidates
MAX_POSSIBLE_CANDIDATES: int = 1000

# HTML_CLEANER config
# https://lxml.de/api/lxml.html.clean.Cleaner-class.html
# https://lxml.de/apidoc/lxml.html.clean.html
HTML_CLEANER: Cleaner = Cleaner(
    annoying_tags=False,
    comments=False,
    embedded=True,  # affects recall?
    forms=False,
    frames=True,
    javascript=False,
    links=False,
    meta=False,
    page_structure=True,
    processing_instructions=False,
    remove_unknown_tags=False,
    safe_attrs_only=False,
    scripts=False,
    style=False,
    kill_tags=[
        "applet",
        "audio",
        "canvas",
        "datalist",
        "embed",
        "figure",
        "label",
        "map",
        "math",
        "object",
        "picture",
        "rdf",
        "svg",
        "video",
    ],
)



# pylint:disable-msg=E0611,I1101
"""
Filters for date parsing and date validators.
"""

## This file is available from https://github.com/adbar/htmldate
## under GNU GPL v3 license

# standard
import logging

from collections import Counter
from datetime import datetime
from functools import lru_cache
from time import mktime
from typing import Match, Optional, Pattern, Union, Counter as Counter_Type

from .settings import CACHE_SIZE, MIN_DATE


LOGGER = logging.getLogger(__name__)
LOGGER.debug("minimum date setting: %s", MIN_DATE)


@lru_cache(maxsize=CACHE_SIZE)
def date_validator(
    date_input: Optional[Union[datetime, str]],
    outputformat: str,
    earliest: datetime,
    latest: datetime,
) -> bool:
    """Validate a string w.r.t. the chosen outputformat and basic heuristics"""
    # safety check
    if date_input is None:
        return False
    # try if date can be parsed using chosen outputformat
    if not isinstance(date_input, datetime):
        # speed-up
        try:
            if outputformat == "%Y-%m-%d":
                dateobject = datetime(
                    int(date_input[:4]), int(date_input[5:7]), int(date_input[8:10])
                )
            # default
            else:
                dateobject = datetime.strptime(date_input, outputformat)
        except ValueError:
            return False
    else:
        dateobject = date_input
    # basic year validation
    year = int(datetime.strftime(dateobject, "%Y"))
    min_year, max_year = earliest.year, latest.year
    # full validation: not newer than today or stored variable
    if (
        min_year <= year <= max_year
        and earliest.timestamp() <= dateobject.timestamp() <= latest.timestamp()
    ):
        return True
    LOGGER.debug("date not valid: %s", date_input)
    return False


@lru_cache(maxsize=16)
def output_format_validator(outputformat: str) -> bool:
    """Validate the output format in the settings"""
    # test with date object
    dateobject = datetime(2017, 9, 1, 0, 0)
    try:
        dateobject.strftime(outputformat)
    # other than ValueError: Python < 3.7 only
    except (NameError, TypeError, ValueError) as err:
        LOGGER.error("wrong output format or type: %s %s", outputformat, err)
        return False
    else:
        # test in abstracto
        if not isinstance(outputformat, str) or "%" not in outputformat:
            LOGGER.error("malformed output format: %s", outputformat)
            return False
    return True


def plausible_year_filter(
    htmlstring: str,
    *,
    pattern: Pattern[str],
    yearpat: Pattern[str],
    earliest: datetime,
    latest: datetime,
    incomplete: bool = False,
) -> Counter_Type[str]:
    """Filter the date patterns to find plausible years only"""
    # slow!
    occurrences = Counter(pattern.findall(htmlstring))
    toremove = set()
    # LOGGER.debug('occurrences: %s', occurrences)
    # look for implausible dates
    for item in occurrences.keys():
        # scrap implausible dates
        year_match = yearpat.search(item)
        if year_match is not None:
            if not incomplete:
                potential_year = int(year_match[1])
            else:
                lastdigits = year_match[1]
                if lastdigits[0] == "9":
                    potential_year = int("19" + lastdigits)
                else:
                    potential_year = int("20" + lastdigits)
            if not earliest.year <= potential_year <= latest.year:
                LOGGER.debug("no potential year: %s", item)
                toremove.add(item)
            # occurrences.remove(item)
            # continue
        else:
            LOGGER.debug("not a year pattern: %s", item)
            toremove.add(item)
    # remove candidates
    for item in toremove:
        del occurrences[item]
    return occurrences


def compare_values(
    reference: int, attempt: str, outputformat: str, original_date: bool
) -> int:
    """Compare the date expression to a reference"""
    try:
        timestamp = int(mktime(datetime.strptime(attempt, outputformat).timetuple()))
    except Exception as err:
        LOGGER.debug("datetime.strptime exception: %s for string %s", err, attempt)
        return reference
    if original_date and (reference == 0 or timestamp < reference):
        reference = timestamp
    elif not original_date and timestamp > reference:
        reference = timestamp
    return reference


@lru_cache(maxsize=CACHE_SIZE)
def filter_ymd_candidate(
    bestmatch: Match[str],
    pattern: Pattern[str],
    original_date: bool,
    copyear: int,
    outputformat: str,
    min_date: datetime,
    max_date: datetime,
) -> Optional[str]:
    """Filter free text candidates in the YMD format"""
    if bestmatch is not None:
        pagedate = "-".join([bestmatch[1], bestmatch[2], bestmatch[3]])
        if date_validator(
            pagedate, "%Y-%m-%d", earliest=min_date, latest=max_date
        ) is True and (copyear == 0 or int(bestmatch[1]) >= copyear):
            LOGGER.debug('date found for pattern "%s": %s', pattern, pagedate)
            return convert_date(pagedate, "%Y-%m-%d", outputformat)
            ## TODO: test and improve
            # if original_date is True:
            #    if copyear == 0 or int(bestmatch[1]) <= copyear:
            #        LOGGER.debug('date found for pattern "%s": %s', pattern, pagedate)
            #        return convert_date(pagedate, '%Y-%m-%d', outputformat)
            # else:
            #    if copyear == 0 or int(bestmatch[1]) >= copyear:
            #        LOGGER.debug('date found for pattern "%s": %s', pattern, pagedate)
            #        return convert_date(pagedate, '%Y-%m-%d', outputformat)
    return None


def convert_date(datestring: str, inputformat: str, outputformat: str) -> str:
    """Parse date and return string in desired format"""
    # speed-up (%Y-%m-%d)
    if inputformat == outputformat:
        return datestring
    # date object (speedup)
    if isinstance(datestring, datetime):
        return datestring.strftime(outputformat)
    # normal
    dateobject = datetime.strptime(datestring, inputformat)
    return dateobject.strftime(outputformat)


def check_extracted_reference(
    reference: int, outputformat: str, min_date: datetime, max_date: datetime
) -> Optional[str]:
    """Test if the extracted reference date can be returned"""
    if reference > 0:
        dateobject = datetime.fromtimestamp(reference)
        converted = dateobject.strftime(outputformat)
        if (
            date_validator(converted, outputformat, earliest=min_date, latest=max_date)
            is True
        ):
            return converted
    return None


def get_min_date(min_date: Optional[Union[datetime, str]]) -> datetime:
    """Validates the minimum date and/or defaults to earliest plausible date"""
    if min_date is not None and isinstance(min_date, str):
        try:
            # internal conversion from Y-M-D format
            min_date = datetime(
                int(min_date[:4]), int(min_date[5:7]), int(min_date[8:10])
            )
        except ValueError:
            min_date = MIN_DATE
    else:
        min_date = MIN_DATE
    return min_date


def get_max_date(max_date: Optional[Union[datetime, str]]) -> datetime:
    """Validates the maximum date and/or defaults to latest plausible date"""
    if max_date is not None and isinstance(max_date, str):
        try:
            # internal conversion from Y-M-D format
            max_date = datetime(
                int(max_date[:4]), int(max_date[5:7]), int(max_date[8:10])
            )
        except ValueError:
            max_date = datetime.now()
    else:
        max_date = datetime.now()
    return max_date



"""
Meta-functions to be applied module-wide.
"""

import logging

from .core import compare_reference
from .extractors import try_date_expr
from .validators import date_validator, filter_ymd_candidate, output_format_validator


LOGGER = logging.getLogger(__name__)


try:
    from charset_normalizer.cd import encoding_languages
    from charset_normalizer.md import is_suspiciously_successive_range
    from charset_normalizer.utils import is_accentuated
# prevent possible changes in function names
except ImportError:
    LOGGER.error("impossible to import charset function name")


def reset_caches() -> None:
    """Reset all known LRU caches used to speed-up processing.
    This may release some memory."""
    # htmldate
    compare_reference.cache_clear()
    date_validator.cache_clear()
    filter_ymd_candidate.cache_clear()
    output_format_validator.cache_clear()
    try_date_expr.cache_clear()
    # charset_normalizer
    try:
        encoding_languages.cache_clear()
        is_suspiciously_successive_range.cache_clear()
        is_accentuated.cache_clear()
    # prevent possible changes in function names
    except (AttributeError, NameError) as err:  # pragma: no cover
        LOGGER.error("impossible to clear cache for function: %s", err)



"""
Seamlessly extract the date of web pages based on URL, header or body.
http://github.com/adbar/htmldate
"""

import re
import sys

from pathlib import Path
from setuptools import setup


# some problems with installation solved this way
extras = {
    "speed": [
        "cchardet >= 2.1.7; python_version < '3.11'",  # build issue
        "dateparser == 1.1.2",  # 1.1.3+ slower
        "faust-cchardet >= 2.1.14; python_version >= '3.11'",  # fix for build
        "urllib3[brotli]",
    ],
}
extras["all"] = extras["speed"]


def get_long_description():
    "Return the README"
    with open("README.rst", "r", encoding="utf-8") as filehandle:
        long_description = filehandle.read()
    # long_description += "\n\n"
    # with open("CHANGELOG.md", encoding="utf8") as f:
    #    long_description += f.read()
    return long_description


def get_version(package):
    "Return package version as listed in `__version__` in `init.py`"
    initfile = Path(package, "__init__.py").read_text()  # Python >= 3.5
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", initfile)[1]


# add argument to compile with mypyc
if len(sys.argv) > 1 and sys.argv[1] == "--use-mypyc":
    sys.argv.pop(1)
    USE_MYPYC = True
    from mypyc.build import mypycify

    ext_modules = mypycify(
        [
            "htmldate/__init__.py",
            "htmldate/core.py",
            "htmldate/extractors.py",
            "htmldate/settings.py",
            "htmldate/utils.py",
            "htmldate/validators.py",
        ],
        opt_level="3",
        multi_file=True,
    )
else:
    ext_modules = []


setup(
    name="htmldate",
    version=get_version("htmldate"),
    description="Fast and robust extraction of original and updated publication dates from URLs and web pages.",
    long_description=get_long_description(),
    classifiers=[
        # As from http://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 5 - Production/Stable",
        # 'Development Status :: 6 - Mature',
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Text Processing :: Linguistic",
        "Topic :: Text Processing :: Markup :: HTML",
    ],
    keywords=[
        "datetime",
        "date-parser",
        "entity-extraction",
        "html-extraction",
        "html-parsing",
        "metadata-extraction",
        "webarchives",
        "web-scraping",
    ],
    url="https://htmldate.readthedocs.io",
    project_urls={
        "Source": "https://github.com/adbar/htmldate",
        "Tracker": "https://github.com/adbar/htmldate/issues",
        "Blog": "https://adrien.barbaresi.eu/blog/tag/htmldate.html",
    },
    author="Adrien Barbaresi",
    author_email="barbaresi@bbaw.de",
    license="GPLv3+",
    packages=["htmldate"],
    include_package_data=True,
    python_requires=">=3.6",
    install_requires=[
        "backports-datetime-fromisoformat; python_version < '3.7'",
        "charset_normalizer >= 3.0.1",
        "dateparser >= 1.1.2",  # 1.1.3+ slower
        "lxml >= 4.9.2",
        "python-dateutil >= 2.8.2",
        "urllib3 >= 1.26, <2",
    ],
    extras_require=extras,
    entry_points={
        "console_scripts": ["htmldate=htmldate.cli:main"],
    },
    # platforms='any',
    tests_require=["pytest"],
    zip_safe=False,
    # optional use of mypyc
    ext_modules=ext_modules,
)



# pylint:disable-msg=W1401
"""
Unit tests for the htmldate library.
"""


import datetime
import io
import logging
import os
import re
import sys

from collections import Counter
from contextlib import redirect_stdout
from unittest.mock import Mock, patch

import pytest

try:
    import dateparser

    EXT_PARSER = True
    PARSER = dateparser.DateDataParser(
        languages=["de", "en"],
        settings={
            "PREFER_DAY_OF_MONTH": "first",
            "PREFER_DATES_FROM": "past",
            "DATE_ORDER": "DMY",
        },
    )  # allow_redetect_language=False,
except ImportError:
    EXT_PARSER = False

from lxml import html
from lxml.etree import XPathEvalError

from htmldate.cli import examine, main, parse_args, process_args
from htmldate.core import (
    compare_reference,
    examine_date_elements,
    find_date,
    search_page,
    search_pattern,
    select_candidate,
)
from htmldate.extractors import (
    custom_parse,
    discard_unwanted,
    external_date_parser,
    extract_partial_url_date,
    regex_parse,
    try_date_expr,
)
from htmldate.meta import reset_caches
from htmldate.settings import MIN_DATE
from htmldate.utils import (
    decode_response,
    fetch_url,
    is_dubious_html,
    load_html,
    strip_faulty_doctypes,
)
from htmldate.validators import (
    convert_date,
    date_validator,
    get_max_date,
    get_min_date,
    output_format_validator,
)


TEST_DIR = os.path.abspath(os.path.dirname(__file__))
OUTPUTFORMAT = "%Y-%m-%d"

LATEST_POSSIBLE = datetime.datetime.now()

logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)


def test_input():
    """test if loaded strings/trees are handled properly"""
    assert is_dubious_html("This is a string.") is True
    htmlstring = "<!DOCTYPE html PUBLIC />\n<html/>"
    beginning = htmlstring[:50].lower()
    assert strip_faulty_doctypes(htmlstring, beginning) == "\n<html/>"
    htmlstring = "<html>\n</html>"
    beginning = htmlstring[:50].lower()
    assert strip_faulty_doctypes(htmlstring, beginning) == htmlstring
    with pytest.raises(TypeError) as err:
        assert load_html(123) is None
    assert "incompatible" in str(err.value)
    assert load_html("<" * 100) is None
    assert load_html("<html><body>XYZ</body></html>") is not None
    assert load_html(b"<html><body>XYZ</body></html>") is not None
    assert load_html(b"<" * 100) is None
    assert load_html(b"<html><body>\x2f\x2e\x9f</body></html>") is not None
    assert (
        load_html("<html><body>\x2f\x2e\x9f</body></html>".encode("latin-1"))
        is not None
    )
    assert load_html("https://httpbin.org/html") is not None
    # encoding declaration
    assert (
        load_html(
            '<?xml version="1.0" encoding="utf-8"?><!DOCTYPE html><html lang="en"><head><meta charset="utf-8"/><meta http-equiv="Content-Type" content="text/html; charset=utf-8"/></head><body><p>A</p><p>B</p></body></html>'
        )
        is not None
    )
    # response decoding
    assert decode_response(b"\x1f\x8babcdef") is not None
    mock = Mock()
    mock.data = b" "
    assert decode_response(mock) is not None
    # find_date logic
    with pytest.raises(TypeError):
        assert find_date(None) is None
    assert find_date("<" * 100) is None
    assert find_date("<html></html>", verbose=True) is None
    assert find_date("<html><body>\u0008this\xdf\n\u001f+\uffff</body></html>") is None
    # min and max date output
    assert get_min_date("2020-02-20").date() == datetime.date(2020, 2, 20)
    assert get_min_date(None).date() == datetime.date(1995, 1, 1)
    assert get_min_date("3030-30-50").date() == datetime.date(1995, 1, 1)
    assert get_max_date("2020-02-20").date() == datetime.date(2020, 2, 20)
    assert get_max_date(None).date() == datetime.date.today()
    assert get_max_date("3030-30-50").date() == datetime.date.today()


def test_sanity():
    """Test if function arguments are interpreted and processed correctly."""
    # XPath looking for date elements
    mytree = html.fromstring("<html><body><p>Test.</p></body></html>")
    with pytest.raises(XPathEvalError):
        examine_date_elements(
            mytree, ".//[Error", OUTPUTFORMAT, True, MIN_DATE, LATEST_POSSIBLE
        )
    result = examine_date_elements(
        mytree, ".//p", OUTPUTFORMAT, True, MIN_DATE, LATEST_POSSIBLE
    )
    assert result is None
    mytree = html.fromstring("<html><body><p>1999/03/05</p></body></html>")
    result = examine_date_elements(
        mytree, ".//p", OUTPUTFORMAT, True, MIN_DATE, LATEST_POSSIBLE
    )
    assert result is not None
    # wrong field values in output format
    assert output_format_validator("%Y-%m-%d") is True
    assert output_format_validator("%M-%Y") is True
    assert output_format_validator("ABC") is False
    assert output_format_validator(123) is False
    assert output_format_validator(("a", "b")) is False
    # assert output_format_validator('%\xaa') is False
    _, discarded = discard_unwanted(
        html.fromstring(
            '<html><body><div id="wm-ipp">000</div><div>AAA</div></body></html>'
        )
    )
    assert len(discarded) == 1
    # reset caches: examine_date_elements used above
    old_values = try_date_expr.cache_info()
    reset_caches()
    assert try_date_expr.cache_info() != old_values


def test_no_date():
    """These pages should not return any date"""
    assert find_date("<html><body>XYZ</body></html>", outputformat="123") is None
    assert (
        find_date(
            "<html><body>XYZ</body></html>", url="http://www.website.com/9999/01/43/"
        )
        is None
    )


def test_exact_date():
    """These pages should return an exact date"""
    ## HTML tree
    assert (
        find_date(
            '<html><head><meta property="dc:created" content="2017-09-01"/></head><body></body></html>'
        )
        == "2017-09-01"
    )
    assert (
        find_date(
            '<html><head><meta property="og:published_time" content="2017-09-01"/></head><body></body></html>',
            original_date=True,
        )
        == "2017-09-01"
    )
    assert (
        find_date(
            '<html><head><meta name="last-modified" content="2017-09-01"/></head><body></body></html>',
            original_date=False,
        )
        == "2017-09-01"
    )
    assert (
        find_date(
            '<html><head><meta property="OG:Updated_Time" content="2017-09-01"/></head><body></body></html>',
            extensive_search=False,
        )
        == "2017-09-01"
    )
    assert (
        find_date(
            '<html><head><Meta Property="og:updated_time" content="2017-09-01"/></head><body></body></html>',
            extensive_search=False,
        )
        == "2017-09-01"
    )
    assert (
        find_date(
            '<html><head><meta name="created" content="2017-01-09"/></head><body></body></html>'
        )
        == "2017-01-09"
    )
    assert (
        find_date(
            '<html><head><meta itemprop="copyrightyear" content="2017"/></head><body></body></html>'
        )
        == "2017-01-01"
    )

    # original date
    assert (
        find_date(
            '<html><head><meta property="OG:Updated_Time" content="2017-09-01"/><meta property="OG:Original_Time" content="2017-07-02"/></head><body></body></html>',
            original_date=True,
        )
        == "2017-07-02"
    )
    assert (
        find_date(
            '<html><head><meta property="OG:Updated_Time" content="2017-09-01"/><meta property="OG:Original_Time" content="2017-07-02"/></head><body></body></html>',
            original_date=False,
        )
        == "2017-09-01"
    )

    ## meta in header
    assert find_date("<html><head><meta/></head><body></body></html>") is None

    assert (
        find_date(
            '<html><head><meta name="og:url" content="http://www.example.com/2018/02/01/entrytitle"/></head><body></body></html>'
        )
        == "2018-02-01"
    )
    assert (
        find_date(
            '<html><head><meta itemprop="datecreated" datetime="2018-02-02"/></head><body></body></html>'
        )
        == "2018-02-02"
    )
    assert (
        find_date(
            '<html><head><meta itemprop="datemodified" content="2018-02-04"/></head><body></body></html>'
        )
        == "2018-02-04"
    )
    assert (
        find_date(
            '<html><head><meta http-equiv="last-modified" content="2018-02-05"/></head><body></body></html>'
        )
        == "2018-02-05"
    )
    assert (
        find_date(
            '<html><head><meta name="lastmodified" content="2018-02-05"/></head><body></body></html>',
            original_date=True,
        )
        == "2018-02-05"
    )
    assert (
        find_date(
            '<html><head><meta name="lastmodified" content="2018-02-05"/></head><body></body></html>',
            original_date=False,
        )
        == "2018-02-05"
    )
    assert (
        find_date(
            '<html><head><meta http-equiv="date" content="2017-09-01"/></head><body></body></html>',
            original_date=True,
        )
        == "2017-09-01"
    )
    assert (
        find_date(
            '<html><head><meta http-equiv="last-modified" content="2018-10-01"/><meta http-equiv="date" content="2017-09-01"/></head><body></body></html>',
            original_date=True,
        )
        == "2017-09-01"
    )
    assert (
        find_date(
            '<html><head><meta http-equiv="last-modified" content="2018-10-01"/><meta http-equiv="date" content="2017-09-01"/></head><body></body></html>',
            original_date=False,
        )
        == "2018-10-01"
    )
    assert (
        find_date(
            '<html><head><meta http-equiv="date" content="2017-09-01"/><meta http-equiv="last-modified" content="2018-10-01"/></head><body></body></html>',
            original_date=True,
        )
        == "2017-09-01"
    )
    assert (
        find_date(
            '<html><head><meta http-equiv="date" content="2017-09-01"/><meta http-equiv="last-modified" content="2018-10-01"/></head><body></body></html>',
            original_date=False,
        )
        == "2018-10-01"
    )
    assert (
        find_date(
            '<html><head><meta name="Publish_Date" content="02.02.2004"/></head><body></body></html>'
        )
        == "2004-02-02"
    )
    assert (
        find_date(
            '<html><head><meta name="pubDate" content="2018-02-06"/></head><body></body></html>'
        )
        == "2018-02-06"
    )
    assert (
        find_date(
            '<html><head><meta pubdate="pubDate" content="2018-02-06"/></head><body></body></html>'
        )
        == "2018-02-06"
    )
    assert (
        find_date(
            '<html><head><meta itemprop="DateModified" datetime="2018-02-06"/></head><body></body></html>'
        )
        == "2018-02-06"
    )
    assert (
        find_date(
            '<html><head><meta name="DC.Issued" content="2021-07-13"/></head><body></body></html>'
        )
        == "2021-07-13"
    )
    assert (
        find_date(
            '<html><head><meta itemprop="dateUpdate" datetime="2018-02-06"/></head><body></body></html>',
            original_date=True,
        )
        == "2018-02-06"
    )
    assert (
        find_date(
            '<html><head><meta itemprop="dateUpdate" datetime="2018-02-06"/></head><body></body></html>',
            original_date=False,
        )
        == "2018-02-06"
    )

    ## time in document body
    assert find_date('<html><body><time datetime="08:00"></body></html>') is None
    assert (
        find_date('<html><body><time datetime="2014-07-10 08:30:45.687"></body></html>')
        == "2014-07-10"
    )
    assert (
        find_date(
            '<html><head></head><body><time class="entry-time" itemprop="datePublished" datetime="2018-04-18T09:57:38+00:00"></body></html>'
        )
        == "2018-04-18"
    )
    assert (
        find_date(
            '<html><body><footer class="article-footer"><p class="byline">Veröffentlicht am <time class="updated" datetime="2019-01-03T14:56:51+00:00">3. Januar 2019 um 14:56 Uhr.</time></p></footer></body></html>'
        )
        == "2019-01-03"
    )
    assert (
        find_date(
            '<html><body><footer class="article-footer"><p class="byline">Veröffentlicht am <time class="updated" datetime="2019-01-03T14:56:51+00:00"></time></p></footer></body></html>'
        )
        == "2019-01-03"
    )
    assert (
        find_date(
            '<html><body><time datetime="2011-09-27" class="entry-date"></time><time datetime="2011-09-28" class="updated"></time></body></html>',
            original_date=True,
        )
        == "2011-09-27"
    )
    # problem here:
    # assert find_date('<html><body><time datetime="2011-09-27" class="entry-date"></time><time datetime="2011-09-28" class="updated"></time></body></html>', original_date=False) == '2011-09-28'
    # assert find_date('<html><body><time datetime="2011-09-28" class="updated"></time><time datetime="2011-09-27" class="entry-date"></time></body></html>', original_date=True) == '2011-09-27'
    assert (
        find_date(
            '<html><body><time datetime="2011-09-28" class="updated"></time><time datetime="2011-09-27" class="entry-date"></time></body></html>',
            original_date=False,
        )
        == "2011-09-28"
    )
    # removed from HTML5 https://www.w3schools.com/TAgs/att_time_datetime_pubdate.asp
    assert (
        find_date(
            '<html><body><time datetime="2011-09-28" pubdate="pubdate"></time></body></html>',
            original_date=False,
        )
        == "2011-09-28"
    )
    assert (
        find_date(
            '<html><body><time datetime="2011-09-28" pubdate="pubdate"></time></body></html>',
            original_date=True,
        )
        == "2011-09-28"
    )
    assert (
        find_date(
            '<html><body><time datetime="2011-09-28" class="entry-date"></time></body></html>',
            original_date=True,
        )
        == "2011-09-28"
    )
    # bug #54
    assert (
        find_date(
            '<html><body><time class="Feed-module--feed__item-meta-time--3t1fg" dateTime="November 29, 2020">November 2020</time></body></html>',
            outputformat="%Y-%m-%d %H:%m:%S",
        )
        == "2020-11-29 00:11:00"  # 00:11 unclear
    )
    assert (
        find_date(
            '<html><body><time class="Feed-module--feed__item-meta-time--3t1fg" dateTime="November 29, 2020">November 2020</time></body></html>',
            outputformat="%Y-%m-%d",
        )
        == "2020-11-29"
    )

    ## precise pattern in document body
    assert (
        find_date(
            '<html><body><font size="2" face="Arial,Geneva,Helvetica">Bei <a href="../../sonstiges/anfrage.php"><b>Bestellungen</b></a> bitte Angabe der Titelnummer nicht vergessen!<br><br>Stand: 03.04.2019</font></body></html>'
        )
        == "2019-04-03"
    )
    assert (
        find_date(
            "<html><body><div>Erstausstrahlung: 01.01.2020</div><div>Preisstand: 03.02.2022 03:00 GMT+1</div></body></html>"
        )
        == "2022-02-03"
    )
    assert find_date("<html><body>Datum: 10.11.2017</body></html>") == "2017-11-10"

    # abbr in document body
    assert (
        find_date(
            '<html><body><abbr class="published">am 12.11.16</abbr></body></html>',
            original_date=False,
        )
        == "2016-11-12"
    )
    assert (
        find_date(
            '<html><body><abbr class="published">am 12.11.16</abbr></body></html>',
            original_date=True,
        )
        == "2016-11-12"
    )
    assert (
        find_date(
            '<html><body><abbr class="published" title="2016-11-12">XYZ</abbr></body></html>',
            original_date=True,
        )
        == "2016-11-12"
    )
    assert (
        find_date(
            '<html><body><abbr class="date-published">8.11.2016</abbr></body></html>'
        )
        == "2016-11-08"
    )
    # valid vs. invalid data-utime
    assert (
        find_date(
            '<html><body><abbr data-utime="1438091078" class="something">A date</abbr></body></html>'
        )
        == "2015-07-28"
    )
    assert (
        find_date(
            '<html><body><abbr data-utime="143809-1078" class="something">A date</abbr></body></html>'
        )
        is None
    )

    # time in document body
    assert (
        find_date("<html><body><time>2018-01-04</time></body></html>") == "2018-01-04"
    )

    # additional XPath expressions
    assert (
        find_date('<html><body><div class="fecha">2018-01-04</div></body></html>')
        == "2018-01-04"
    )

    ## other expressions in document body
    assert (
        find_date('<html><body>"datePublished":"2018-01-04"</body></html>')
        == "2018-01-04"
    )
    assert find_date("<html><body>Stand: 1.4.18</body></html>") == "2018-04-01"

    # free text
    assert find_date("<html><body>&copy; 2017</body></html>") == "2017-01-01"
    assert find_date("<html><body>© 2017</body></html>") == "2017-01-01"
    assert (
        find_date(
            "<html><body><p>Dieses Datum ist leider ungültig: 30. Februar 2018.</p></body></html>",
            extensive_search=False,
        )
        is None
    )
    assert (
        find_date(
            "<html><body><p>Dieses Datum ist leider ungültig: 30. Februar 2018.</p></body></html>"
        )
        == "2018-01-01"
    )

    # URL in IMG
    # header
    assert (
        find_date(
            '<html><meta property="og:image" content="https://example.org/img/2019-05-05/test.jpg"><body></body></html>'
        )
        == "2019-05-05"
    )
    assert (
        find_date(
            '<html><meta property="og:image" content="https://example.org/img/test.jpg"><body></body></html>'
        )
        is None
    )
    # body
    assert (
        find_date(
            '<html><body><img src="https://example.org/img/2019-05-05/test.jpg"/></body></html>'
        )
        == "2019-05-05"
    )
    assert (
        find_date(
            '<html><body><img src="https://example.org/img/test.jpg"/></body></html>'
        )
        is None
    )
    assert (
        find_date(
            '<html><body><img src="https://example.org/img/2019-05-03/test.jpg"/><img src="https://example.org/img/2019-05-04/test.jpg"/><img src="https://example.org/img/2019-05-05/test.jpg"/></body></html>'
        )
        == "2019-05-05"
    )
    assert (
        find_date(
            '<html><body><img src="https://example.org/img/2019-05-05/test.jpg"/><img src="https://example.org/img/2019-05-04/test.jpg"/><img src="https://example.org/img/2019-05-03/test.jpg"/></body></html>'
        )
        == "2019-05-05"
    )
    # in title
    assert (
        find_date(
            "<html><head><title>Bericht zur Coronalage vom 22.04.2020 – worauf wartet die Politik? – DIE ACHSE DES GUTEN. ACHGUT.COM</title></head></html>"
        )
        == "2020-04-22"
    )
    # in unknown div
    assert (
        find_date(
            '<html><body><div class="spip spip-block-right" style="text-align:right;">Le 26 juin 2019</div></body></html>',
            extensive_search=False,
        )
        is None
    )
    assert (
        find_date(
            '<html><body><div class="spip spip-block-right" style="text-align:right;">Le 26 juin 2019</div></body></html>',
            extensive_search=True,
        )
        == "2019-06-26"
    )
    # in link title
    assert (
        find_date(
            '<html><body><a class="ribbon date " title="12th December 2018" href="https://example.org/" itemprop="url">Text</a></body></html>'
        )
        == "2018-12-12"
    )

    # min_date parameter
    assert (
        find_date(
            '<html><meta><meta property="article:published_time" content="1991-01-02T01:01:00+01:00"></meta><body></body></html>',
            min_date="2000-01-01",
        )
        is None
    )
    assert (
        find_date(
            '<html><meta><meta property="article:published_time" content="1991-01-02T01:01:00+01:00"></meta><body></body></html>',
            min_date="1990-01-01",
        )
        == "1991-01-02"
    )

    # wild text in body
    assert (
        find_date("<html><body>Wed, 19 Oct 2022 14:24:05 +0000</body></html>")
        == "2022-10-19"
    )


def test_date_validator():
    """test internal date validation"""
    assert (
        date_validator(
            "2016-01-01", OUTPUTFORMAT, earliest=MIN_DATE, latest=LATEST_POSSIBLE
        )
        is True
    )
    assert (
        date_validator(
            "1998-08-08", OUTPUTFORMAT, earliest=MIN_DATE, latest=LATEST_POSSIBLE
        )
        is True
    )
    assert (
        date_validator(
            "2001-12-31", OUTPUTFORMAT, earliest=MIN_DATE, latest=LATEST_POSSIBLE
        )
        is True
    )
    assert (
        date_validator(
            "1992-07-30", OUTPUTFORMAT, earliest=MIN_DATE, latest=LATEST_POSSIBLE
        )
        is False
    )
    assert (
        date_validator(
            "1901-13-98", OUTPUTFORMAT, earliest=MIN_DATE, latest=LATEST_POSSIBLE
        )
        is False
    )
    assert (
        date_validator(
            "202-01", OUTPUTFORMAT, earliest=MIN_DATE, latest=LATEST_POSSIBLE
        )
        is False
    )
    assert (
        date_validator("1922", "%Y", earliest=MIN_DATE, latest=LATEST_POSSIBLE) is False
    )
    assert (
        date_validator("2004", "%Y", earliest=MIN_DATE, latest=LATEST_POSSIBLE) is True
    )
    assert (
        date_validator(
            "1991-01-02",
            OUTPUTFORMAT,
            earliest=datetime.datetime(1990, 1, 1),
            latest=LATEST_POSSIBLE,
        )
        is True
    )
    assert (
        date_validator(
            "1991-01-02",
            OUTPUTFORMAT,
            earliest=datetime.datetime(1992, 1, 1),
            latest=LATEST_POSSIBLE,
        )
        is False
    )
    assert (
        date_validator(
            "1991-01-02",
            OUTPUTFORMAT,
            earliest=MIN_DATE,
            latest=datetime.datetime(1990, 1, 1),
        )
        is False
    )
    assert (
        date_validator(
            "1991-01-02",
            OUTPUTFORMAT,
            earliest=datetime.datetime(1990, 1, 1),
            latest=datetime.datetime(1995, 1, 1),
        )
        is True
    )
    assert (
        date_validator(
            "1991-01-02",
            OUTPUTFORMAT,
            earliest=datetime.datetime(1990, 1, 1),
            latest=datetime.datetime(1990, 12, 31),
        )
        is False
    )


def test_convert_date():
    """test date conversion"""
    assert convert_date("2016-11-18", "%Y-%m-%d", "%d %B %Y") == "18 November 2016"
    assert convert_date("18 November 2016", "%d %B %Y", "%Y-%m-%d") == "2016-11-18"
    dateobject = datetime.datetime.strptime("2016-11-18", "%Y-%m-%d")
    assert convert_date(dateobject, "%d %B %Y", "%Y-%m-%d") == "2016-11-18"


def test_try_date_expr():
    """test date extraction via external package"""
    find_date.extensive_search = False
    assert (
        try_date_expr(
            "Fri, Sept 1, 2017", OUTPUTFORMAT, False, MIN_DATE, LATEST_POSSIBLE
        )
        is None
    )
    find_date.extensive_search = True
    assert (
        try_date_expr(
            "Friday, September 01, 2017", OUTPUTFORMAT, True, MIN_DATE, LATEST_POSSIBLE
        )
        == "2017-09-01"
    )
    assert (
        try_date_expr(
            "Fr, 1 Sep 2017 16:27:51 MESZ",
            OUTPUTFORMAT,
            True,
            MIN_DATE,
            LATEST_POSSIBLE,
        )
        == "2017-09-01"
    )
    assert (
        try_date_expr(
            "Freitag, 01. September 2017", OUTPUTFORMAT, True, MIN_DATE, LATEST_POSSIBLE
        )
        == "2017-09-01"
    )
    assert (
        try_date_expr(
            "Am 1. September 2017 um 15:36 Uhr schrieb",
            OUTPUTFORMAT,
            True,
            MIN_DATE,
            LATEST_POSSIBLE,
        )
        == "2017-09-01"
    )
    assert (
        try_date_expr(
            "Fri - September 1 - 2017", OUTPUTFORMAT, True, MIN_DATE, LATEST_POSSIBLE
        )
        == "2017-09-01"
    )
    assert (
        try_date_expr("1.9.2017", OUTPUTFORMAT, True, MIN_DATE, LATEST_POSSIBLE)
        == "2017-09-01"
    )
    assert (
        try_date_expr("1/9/17", OUTPUTFORMAT, True, MIN_DATE, LATEST_POSSIBLE)
        == "2017-09-01"
    )  # assuming MDY format
    assert (
        try_date_expr("201709011234", OUTPUTFORMAT, True, MIN_DATE, LATEST_POSSIBLE)
        == "2017-09-01"
    )
    # other output format
    assert (
        try_date_expr("1.9.2017", "%d %B %Y", True, MIN_DATE, LATEST_POSSIBLE)
        == "01 September 2017"
    )
    # wrong
    assert try_date_expr("201", OUTPUTFORMAT, True, MIN_DATE, LATEST_POSSIBLE) is None
    assert (
        try_date_expr("14:35:10", OUTPUTFORMAT, True, MIN_DATE, LATEST_POSSIBLE) is None
    )
    assert (
        try_date_expr("12:00 h", OUTPUTFORMAT, True, MIN_DATE, LATEST_POSSIBLE) is None
    )
    # date range
    assert (
        try_date_expr("2005-2006", OUTPUTFORMAT, True, MIN_DATE, LATEST_POSSIBLE)
        is None
    )
    # Mandarin
    assert (
        try_date_expr(
            "发布时间: 2022-02-25 14:34", OUTPUTFORMAT, True, MIN_DATE, LATEST_POSSIBLE
        )
        == "2022-02-25"
    )


# def test_header():
#    assert examine_header(tree, OUTPUTFORMAT, PARSER)


def test_compare_reference(
    extensive_search=False,
    original_date=False,
    min_date=MIN_DATE,
    max_date=LATEST_POSSIBLE,
):
    """test comparison function"""
    assert (
        compare_reference(
            0, "AAAA", OUTPUTFORMAT, extensive_search, original_date, min_date, max_date
        )
        == 0
    )
    assert (
        compare_reference(
            1517500000,
            "2018-33-01",
            OUTPUTFORMAT,
            extensive_search,
            original_date,
            min_date,
            max_date,
        )
        == 1517500000
    )
    assert (
        1517400000
        < compare_reference(
            0,
            "2018-02-01",
            OUTPUTFORMAT,
            extensive_search,
            original_date,
            min_date,
            max_date,
        )
        < 1517500000
    )
    assert (
        compare_reference(
            1517500000,
            "2018-02-01",
            OUTPUTFORMAT,
            extensive_search,
            original_date,
            min_date,
            max_date,
        )
        == 1517500000
    )


def test_candidate_selection(min_date=MIN_DATE, max_date=LATEST_POSSIBLE):
    """test the algorithm for several candidates"""
    original_date = False
    # patterns
    catch = re.compile(r"([0-9]{4})-([0-9]{2})-([0-9]{2})")
    yearpat = re.compile(r"^([0-9]{4})")
    # nonsense
    occurrences = Counter(
        [
            "20208956",
            "20208956",
            "20208956",
            "19018956",
            "209561",
            "22020895607-12",
            "2-28",
        ]
    )
    result = select_candidate(
        occurrences, catch, yearpat, original_date, min_date, max_date
    )
    assert result is None
    # plausible
    occurrences = Counter(
        [
            "2016-12-23",
            "2016-12-23",
            "2016-12-23",
            "2016-12-23",
            "2017-08-11",
            "2016-07-12",
            "2017-11-28",
        ]
    )
    result = select_candidate(
        occurrences, catch, yearpat, original_date, min_date, max_date
    )
    assert result.group(0) == "2017-11-28"
    original_date = True
    result = select_candidate(
        occurrences, catch, yearpat, original_date, min_date, max_date
    )
    assert result.group(0) == "2016-07-12"
    # mix plausible/implausible
    occurrences = Counter(
        ["2116-12-23", "2116-12-23", "2116-12-23", "2017-08-11", "2017-08-11"]
    )
    result = select_candidate(
        occurrences, catch, yearpat, original_date, min_date, max_date
    )
    assert result.group(0) == "2017-08-11"
    original_date = False
    occurrences = Counter(
        ["2116-12-23", "2116-12-23", "2116-12-23", "2017-08-11", "2017-08-11"]
    )
    result = select_candidate(
        occurrences, catch, yearpat, original_date, min_date, max_date
    )
    assert result.group(0) == "2017-08-11"
    # taking date present twice, corner case
    occurrences = Counter(
        ["2016-12-23", "2016-12-23", "2017-08-11", "2017-08-11", "2017-08-11"]
    )
    result = select_candidate(
        occurrences, catch, yearpat, original_date, min_date, max_date
    )
    assert result.group(0) == "2016-12-23"


def test_regex_parse():
    """test date extraction using rules and regular expressions"""
    assert regex_parse("3. Dezember 2008") is not None
    assert regex_parse("33. Dezember 2008") is None
    assert regex_parse("3 Aralık 2008 Çarşamba") is not None
    assert regex_parse("3 Aralık 2008") is not None
    assert regex_parse("Tuesday, March 26th, 2019") is not None
    assert regex_parse("March 26, 2019") is not None
    assert regex_parse("3rd Tuesday in March") is None
    assert regex_parse("Mart 26, 2019") is not None
    assert regex_parse("Salı, Mart 26, 2019") is not None
    assert regex_parse("36/14/2016") is None
    assert regex_parse("January 36 1998") is None
    assert (
        custom_parse("January 12 1098", OUTPUTFORMAT, MIN_DATE, LATEST_POSSIBLE) is None
    )
    assert custom_parse("1998-01", OUTPUTFORMAT, MIN_DATE, LATEST_POSSIBLE) is not None
    assert custom_parse("10.10.98", OUTPUTFORMAT, MIN_DATE, LATEST_POSSIBLE) is not None
    assert custom_parse("12122004", OUTPUTFORMAT, MIN_DATE, LATEST_POSSIBLE) is None
    assert (
        custom_parse("3/14/2016", OUTPUTFORMAT, MIN_DATE, LATEST_POSSIBLE) is not None
    )
    assert custom_parse("20041212", OUTPUTFORMAT, MIN_DATE, LATEST_POSSIBLE) is not None
    assert custom_parse("20041212", OUTPUTFORMAT, MIN_DATE, LATEST_POSSIBLE) is not None
    assert custom_parse("1212-20-04", OUTPUTFORMAT, MIN_DATE, LATEST_POSSIBLE) is None
    assert custom_parse("1212-20-04", OUTPUTFORMAT, MIN_DATE, LATEST_POSSIBLE) is None
    assert (
        custom_parse("2004-12-12", OUTPUTFORMAT, MIN_DATE, LATEST_POSSIBLE) is not None
    )
    assert (
        custom_parse("2004-12-12", OUTPUTFORMAT, MIN_DATE, LATEST_POSSIBLE) is not None
    )
    assert custom_parse("33.20.2004", OUTPUTFORMAT, MIN_DATE, LATEST_POSSIBLE) is None
    assert (
        custom_parse("12.12.2004", OUTPUTFORMAT, MIN_DATE, LATEST_POSSIBLE) is not None
    )
    assert custom_parse("2019 28 meh", OUTPUTFORMAT, MIN_DATE, LATEST_POSSIBLE) is None
    assert custom_parse("2019 28 meh", OUTPUTFORMAT, MIN_DATE, LATEST_POSSIBLE) is None
    # regex-based matches
    assert (
        custom_parse("abcd 20041212 efgh", OUTPUTFORMAT, MIN_DATE, LATEST_POSSIBLE)
        is not None
    )
    assert (
        custom_parse("abcd 2004-2-12 efgh", OUTPUTFORMAT, MIN_DATE, LATEST_POSSIBLE)
        is not None
    )
    assert (
        custom_parse("abcd 2004-2 efgh", OUTPUTFORMAT, MIN_DATE, LATEST_POSSIBLE)
        is not None
    )
    assert (
        custom_parse("abcd 2004-2 efgh", OUTPUTFORMAT, MIN_DATE, LATEST_POSSIBLE)
        is not None
    )
    assert (
        custom_parse(
            "abcd 32. Januar 2020 efgh", OUTPUTFORMAT, MIN_DATE, LATEST_POSSIBLE
        )
        is None
    )
    # plausible but impossible dates
    assert (
        custom_parse("February 29 2008", OUTPUTFORMAT, MIN_DATE, LATEST_POSSIBLE)
        == "2008-02-29"
    )
    assert (
        custom_parse("February 30 2008", OUTPUTFORMAT, MIN_DATE, LATEST_POSSIBLE)
        is None
    )
    assert (
        custom_parse(
            "XXTag, den 29. Februar 2008", OUTPUTFORMAT, MIN_DATE, LATEST_POSSIBLE
        )
        == "2008-02-29"
    )
    assert (
        custom_parse(
            "XXTag, den 30. Februar 2008", OUTPUTFORMAT, MIN_DATE, LATEST_POSSIBLE
        )
        is None
    )
    # for Nones caused by newlines and duplicates
    assert regex_parse("January 1st, 1998") is not None
    assert regex_parse("February 1st, 1998") is not None
    assert regex_parse("March 1st, 1998") is not None
    assert regex_parse("April 1st, 1998") is not None
    assert regex_parse("May 1st, 1998") is not None
    assert regex_parse("June 1st, 1998") is not None
    assert regex_parse("July 1st, 1998") is not None
    assert regex_parse("August 1st, 1998") is not None
    assert regex_parse("September 1st, 1998") is not None
    assert regex_parse("October 1st, 1998") is not None
    assert regex_parse("November 1st, 1998") is not None
    assert regex_parse("December 1st, 1998") is not None
    assert regex_parse("Jan 1st, 1998") is not None
    assert regex_parse("Feb 1st, 1998") is not None
    assert regex_parse("Mar 1st, 1998") is not None
    assert regex_parse("Apr 1st, 1998") is not None
    assert regex_parse("Jun 1st, 1998") is not None
    assert regex_parse("Jul 1st, 1998") is not None
    assert regex_parse("Aug 1st, 1998") is not None
    assert regex_parse("Sep 1st, 1998") is not None
    assert regex_parse("Oct 1st, 1998") is not None
    assert regex_parse("Nov 1st, 1998") is not None
    assert regex_parse("Dec 1st, 1998") is not None
    assert regex_parse("Januar 1, 1998") is not None
    assert regex_parse("Jänner 1, 1998") is not None
    assert regex_parse("Februar 1, 1998") is not None
    assert regex_parse("Feber 1, 1998") is not None
    assert regex_parse("März 1, 1998") is not None
    assert regex_parse("April 1, 1998") is not None
    assert regex_parse("Mai 1, 1998") is not None
    assert regex_parse("Juni 1, 1998") is not None
    assert regex_parse("Juli 1, 1998") is not None
    assert regex_parse("August 1, 1998") is not None
    assert regex_parse("September 1, 1998") is not None
    assert regex_parse("Oktober 1, 1998") is not None
    assert regex_parse("1. Okt. 1998") is not None
    assert regex_parse("November 1, 1998") is not None
    assert regex_parse("Dezember 1, 1998") is not None
    assert regex_parse("Ocak 1, 1998") is not None
    assert regex_parse("Şubat 1, 1998") is not None
    assert regex_parse("Mart 1, 1998") is not None
    assert regex_parse("Nisan 1, 1998") is not None
    assert regex_parse("Mayıs 1, 1998") is not None
    assert regex_parse("Haziran 1, 1998") is not None
    assert regex_parse("Temmuz 1, 1998") is not None
    assert regex_parse("Ağustos 1, 1998") is not None
    assert regex_parse("Eylül 1, 1998") is not None
    assert regex_parse("Ekim 1, 1998") is not None
    assert regex_parse("Kasım 1, 1998") is not None
    assert regex_parse("Aralık 1, 1998") is not None
    assert regex_parse("Oca 1, 1998") is not None
    assert regex_parse("Şub 1, 1998") is not None
    assert regex_parse("Mar 1, 1998") is not None
    assert regex_parse("Nis 1, 1998") is not None
    assert regex_parse("May 1, 1998") is not None
    assert regex_parse("Haz 1, 1998") is not None
    assert regex_parse("Tem 1, 1998") is not None
    assert regex_parse("Ağu 1, 1998") is not None
    assert regex_parse("Eyl 1, 1998") is not None
    assert regex_parse("Eki 1, 1998") is not None
    assert regex_parse("Kas 1, 1998") is not None
    assert regex_parse("Ara 1, 1998") is not None
    assert regex_parse("1 January 1998") is not None
    assert regex_parse("1 February 1998") is not None
    assert regex_parse("1 March 1998") is not None
    assert regex_parse("1 April 1998") is not None
    assert regex_parse("1 May 1998") is not None
    assert regex_parse("1 June 1998") is not None
    assert regex_parse("1 July 1998") is not None
    assert regex_parse("1 August 1998") is not None
    assert regex_parse("1 September 1998") is not None
    assert regex_parse("1 October 1998") is not None
    assert regex_parse("1 November 1998") is not None
    assert regex_parse("1 December 1998") is not None
    assert regex_parse("1 Jan 1998") is not None
    assert regex_parse("1 Feb 1998") is not None
    assert regex_parse("1 Mar 1998") is not None
    assert regex_parse("1 Apr 1998") is not None
    assert regex_parse("1 Jun 1998") is not None
    assert regex_parse("1 Jul 1998") is not None
    assert regex_parse("1 Aug 1998") is not None
    assert regex_parse("1 Sep 1998") is not None
    assert regex_parse("1 Oct 1998") is not None
    assert regex_parse("1 Nov 1998") is not None
    assert regex_parse("1 Dec 1998") is not None
    assert regex_parse("1 Januar 1998") is not None
    assert regex_parse("1 Jänner 1998") is not None
    assert regex_parse("1 Februar 1998") is not None
    assert regex_parse("1 Feber 1998") is not None
    assert regex_parse("1 März 1998") is not None
    assert regex_parse("1 April 1998") is not None
    assert regex_parse("1 Mai 1998") is not None
    assert regex_parse("1 Juni 1998") is not None
    assert regex_parse("1 Juli 1998") is not None
    assert regex_parse("1 August 1998") is not None
    assert regex_parse("1 September 1998") is not None
    assert regex_parse("1 Oktober 1998") is not None
    assert regex_parse("1 November 1998") is not None
    assert regex_parse("1 Dezember 1998") is not None
    assert regex_parse("1 Ocak 1998") is not None
    assert regex_parse("1 Şubat 1998") is not None
    assert regex_parse("1 Mart 1998") is not None
    assert regex_parse("1 Nisan 1998") is not None
    assert regex_parse("1 Mayıs 1998") is not None
    assert regex_parse("1 Haziran 1998") is not None
    assert regex_parse("1 Temmuz 1998") is not None
    assert regex_parse("1 Ağustos 1998") is not None
    assert regex_parse("1 Eylül 1998") is not None
    assert regex_parse("1 Ekim 1998") is not None
    assert regex_parse("1 Kasım 1998") is not None
    assert regex_parse("1 Aralık 1998") is not None
    assert regex_parse("1 Oca 1998") is not None
    assert regex_parse("1 Şub 1998") is not None
    assert regex_parse("1 Mar 1998") is not None
    assert regex_parse("1 Nis 1998") is not None
    assert regex_parse("1 May 1998") is not None
    assert regex_parse("1 Haz 1998") is not None
    assert regex_parse("1 Tem 1998") is not None
    assert regex_parse("1 Ağu 1998") is not None
    assert regex_parse("1 Eyl 1998") is not None
    assert regex_parse("1 Eki 1998") is not None
    assert regex_parse("1 Kas 1998") is not None
    assert regex_parse("1 Ara 1998") is not None


def test_external_date_parser():
    """test external date parser"""
    assert (
        external_date_parser("Wednesday, January 1st 2020", OUTPUTFORMAT)
        == "2020-01-01"
    )
    assert external_date_parser("Random text with 2020", OUTPUTFORMAT) is None
    # https://github.com/scrapinghub/dateparser/issues/333
    # assert external_date_parser('1 January 0001', '%d %B %Y') in ('01 January 1', '01 January 0001')
    assert external_date_parser("1 January 1900", "%d %B %Y") == "01 January 1900"
    # https://github.com/scrapinghub/dateparser/issues/406
    assert (
        external_date_parser("2018-04-12 17:20:03.12345678999a", OUTPUTFORMAT)
        == "2018-12-04"
    )
    # https://github.com/scrapinghub/dateparser/issues/685
    assert external_date_parser("12345678912 days", OUTPUTFORMAT) is None
    # https://github.com/scrapinghub/dateparser/issues/680
    assert external_date_parser("2.2250738585072011e-308", OUTPUTFORMAT) is None
    assert external_date_parser("⁰⁴⁵₀₁₂", OUTPUTFORMAT) is None


def test_url():
    """test url parameter"""
    assert (
        find_date(
            "<html><body><p>Aaa, bbb.</p></body></html>",
            url="http://example.com/category/2016/07/12/key-words",
        )
        == "2016-07-12"
    )
    assert (
        find_date(
            "<html><body><p>Aaa, bbb.</p></body></html>",
            url="http://example.com/2016/key-words",
        )
        is None
    )
    assert (
        find_date(
            "<html><body><p>Aaa, bbb.</p></body></html>",
            url="http://www.kreditwesen.org/widerstand-berlin/2012-11-29/keine-kurzung-bei-der-jugend-klubs-konnen-vorerst-aufatmen-bvv-beschliest-haushaltsplan/",
        )
        == "2012-11-29"
    )
    assert (
        find_date(
            "<html><body><p>Aaa, bbb.</p></body></html>",
            url="http://www.kreditwesen.org/widerstand-berlin/2012-11/keine-kurzung-bei-der-jugend-klubs-konnen-vorerst-aufatmen-bvv-beschliest-haushaltsplan/",
        )
        == "2012-11-01"
    )
    assert (
        find_date(
            "<html><body><p>Aaa, bbb.</p></body></html>",
            url="http://www.kreditwesen.org/widerstand-berlin/6666-42-87/",
        )
        is None
    )
    assert (
        find_date(
            "<html><body><p>Z.</p></body></html>",
            url="https://www.pamelaandersonfoundation.org/news/2019/6/26/dm4wjh7skxerzzw8qa8cklj8xdri5j",
        )
        == "2019-06-26"
    )
    assert (
        extract_partial_url_date(
            "https://testsite.org/2018/01/test", "%Y-%m-%d", MIN_DATE, LATEST_POSSIBLE
        )
        == "2018-01-01"
    )
    assert (
        extract_partial_url_date(
            "https://testsite.org/2018/33/test", "%Y-%m-%d", MIN_DATE, LATEST_POSSIBLE
        )
        is None
    )


def test_approximate_url():
    """test url parameter"""
    assert (
        find_date(
            "<html><body><p>Aaa, bbb.</p></body></html>",
            url="http://example.com/blog/2016/07/key-words",
        )
        == "2016-07-01"
    )
    assert (
        find_date(
            "<html><body><p>Aaa, bbb.</p></body></html>",
            url="http://example.com/category/2016/",
        )
        is None
    )


def test_search_pattern(
    original_date=False, min_date=MIN_DATE, max_date=LATEST_POSSIBLE
):
    """test pattern search in strings"""
    #
    pattern = re.compile(r"\D([0-9]{4}[/.-][0-9]{2})\D")
    catch = re.compile(r"([0-9]{4})[/.-]([0-9]{2})")
    yearpat = re.compile(r"^([12][0-9]{3})")
    assert (
        search_pattern(
            "It happened on the 202.E.19, the day when it all began.",
            pattern,
            catch,
            yearpat,
            original_date,
            min_date,
            max_date,
        )
        is None
    )
    assert (
        search_pattern(
            "The date is 2002.02.15.",
            pattern,
            catch,
            yearpat,
            original_date,
            min_date,
            max_date,
        )
        is not None
    )
    assert (
        search_pattern(
            "http://www.url.net/index.html",
            pattern,
            catch,
            yearpat,
            original_date,
            min_date,
            max_date,
        )
        is None
    )
    assert (
        search_pattern(
            "http://www.url.net/2016/01/index.html",
            pattern,
            catch,
            yearpat,
            original_date,
            min_date,
            max_date,
        )
        is not None
    )
    #
    pattern = re.compile(r"\D([0-9]{2}[/.-][0-9]{4})\D")
    catch = re.compile(r"([0-9]{2})[/.-]([0-9]{4})")
    yearpat = re.compile(r"([12][0-9]{3})$")
    assert (
        search_pattern(
            "It happened on the 202.E.19, the day when it all began.",
            pattern,
            catch,
            yearpat,
            original_date,
            min_date,
            max_date,
        )
        is None
    )
    assert (
        search_pattern(
            "It happened on the 15.02.2002, the day when it all began.",
            pattern,
            catch,
            yearpat,
            original_date,
            min_date,
            max_date,
        )
        is not None
    )
    #
    pattern = re.compile(r"\D(2[01][0-9]{2})\D")
    catch = re.compile(r"(2[01][0-9]{2})")
    yearpat = re.compile(r"^(2[01][0-9]{2})")
    assert (
        search_pattern(
            "It happened in the film 300.",
            pattern,
            catch,
            yearpat,
            original_date,
            min_date,
            max_date,
        )
        is None
    )
    assert (
        search_pattern(
            "It happened in 2002.",
            pattern,
            catch,
            yearpat,
            original_date,
            min_date,
            max_date,
        )
        is not None
    )


def test_search_html(original_date=False, min_date=MIN_DATE, max_date=LATEST_POSSIBLE):
    "Test the pattern search in raw HTML"
    # tree input
    assert (
        search_page(
            "<html><body><p>The date is 5/2010</p></body></html>",
            OUTPUTFORMAT,
            original_date,
            min_date,
            max_date,
        )
        == "2010-05-01"
    )
    assert (
        search_page(
            "<html><body><p>The date is 5.5.2010</p></body></html>",
            OUTPUTFORMAT,
            original_date,
            min_date,
            max_date,
        )
        == "2010-05-05"
    )
    assert (
        search_page(
            "<html><body><p>The date is 11/10/99</p></body></html>",
            OUTPUTFORMAT,
            original_date,
            min_date,
            max_date,
        )
        == "1999-10-11"
    )
    assert (
        search_page(
            "<html><body><p>The date is 3/3/11</p></body></html>",
            OUTPUTFORMAT,
            original_date,
            min_date,
            max_date,
        )
        == "2011-03-03"
    )
    assert (
        search_page(
            "<html><body><p>The date is 06.12.06</p></body></html>",
            OUTPUTFORMAT,
            original_date,
            min_date,
            max_date,
        )
        == "2006-12-06"
    )
    assert (
        search_page(
            "<html><body><p>The timestamp is 20140915D15:23H</p></body></html>",
            OUTPUTFORMAT,
            original_date,
            min_date,
            max_date,
        )
        == "2014-09-15"
    )
    assert (
        search_page(
            "<html><body><p>It could be 2015-04-30 or 2003-11-24.</p></body></html>",
            OUTPUTFORMAT,
            True,
            min_date,
            max_date,
        )
        == "2003-11-24"
    )
    assert (
        search_page(
            "<html><body><p>It could be 2015-04-30 or 2003-11-24.</p></body></html>",
            OUTPUTFORMAT,
            False,
            min_date,
            max_date,
        )
        == "2015-04-30"
    )
    assert (
        search_page(
            "<html><body><p>It could be 03/03/2077 or 03/03/2013.</p></body></html>",
            OUTPUTFORMAT,
            original_date,
            min_date,
            max_date,
        )
        == "2013-03-03"
    )
    assert (
        search_page(
            "<html><body><p>It could not be 03/03/2077 or 03/03/1988.</p></body></html>",
            OUTPUTFORMAT,
            original_date,
            min_date,
            max_date,
        )
        is None
    )
    assert (
        search_page(
            "<html><body><p>© The Web Association 2013.</p></body></html>",
            OUTPUTFORMAT,
            original_date,
            min_date,
            max_date,
        )
        == "2013-01-01"
    )
    assert (
        search_page(
            "<html><body><p>Next © Copyright 2018</p></body></html>",
            OUTPUTFORMAT,
            original_date,
            min_date,
            max_date,
        )
        == "2018-01-01"
    )
    assert (
        search_page(
            "<html><body><p> © Company 2014-2019 </p></body></html>",
            OUTPUTFORMAT,
            original_date,
            min_date,
            max_date,
        )
        == "2019-01-01"
    )


def test_idiosyncrasies():
    assert (
        find_date("<html><body><p><em>Last updated: 5/5/20</em></p></body></html>")
        == "2020-05-05"
    )
    assert (
        find_date("<html><body><p><em>Last updated: 01/23/2021</em></p></body></html>")
        == "2021-01-23"
    )
    assert (
        find_date("<html><body><p><em>Last updated: 01/23/21</em></p></body></html>")
        == "2021-01-23"
    )
    assert (
        find_date("<html><body><p><em>Last updated: 1/23/21</em></p></body></html>")
        == "2021-01-23"
    )
    assert (
        find_date("<html><body><p><em>Last updated: 23/1/21</em></p></body></html>")
        == "2021-01-23"
    )
    assert (
        find_date("<html><body><p><em>Last updated: 23/01/21</em></p></body></html>")
        == "2021-01-23"
    )
    assert (
        find_date("<html><body><p><em>Last updated: 23/01/2021</em></p></body></html>")
        == "2021-01-23"
    )
    assert (
        find_date("<html><body><p><em>Last updated: 33/23/3033</em></p></body></html>")
        is None
    )
    assert (
        find_date("<html><body><p><em>Published: 5/5/2020</em></p></body></html>")
        == "2020-05-05"
    )
    assert (
        find_date("<html><body><p><em>Published in: 05.05.2020</em></p></body></html>")
        == "2020-05-05"
    )
    assert (
        find_date("<html><body><p><em>Son güncelleme: 5/5/20</em></p></body></html>")
        == "2020-05-05"
    )
    assert (
        find_date("<html><body><p><em>Son güncellenme: 5/5/2020</em></p></body></html>")
        == "2020-05-05"
    )
    assert (
        find_date(
            "<html><body><p><em>Yayımlama tarihi: 05.05.2020</em></p></body></html>"
        )
        == "2020-05-05"
    )
    assert (
        find_date(
            "<html><body><p><em>Son güncelleme tarihi: 5/5/20</em></p></body></html>"
        )
        == "2020-05-05"
    )
    assert (
        find_date(
            "<html><body><p><em>5/5/20 tarihinde güncellendi.</em></p></body></html>"
        )
        == "2020-05-05"
    )
    assert (
        find_date(
            """<html><body><p><em>5/5/20'de güncellendi.</em></p></body></html>"""
        )
        == "2020-05-05"
    )
    assert (
        find_date(
            "<html><body><p><em>5/5/2020 tarihinde yayımlandı.</em></p></body></html>"
        )
        == "2020-05-05"
    )
    assert (
        find_date(
            "<html><body><p><em>05.05.2020 tarihinde yayınlandı.</em></p></body></html>"
        )
        == "2020-05-05"
    )


def test_parser():
    """test argument parsing for the command-line interface"""
    testargs = [
        "-f",
        "-v",
        "--original",
        "-max",
        "2015-12-31",
        "-u",
        "https://www.example.org",
    ]
    with patch.object(sys, "argv", testargs):
        args = parse_args(testargs)
    assert args.fast is True
    assert args.original is True
    assert args.verbose is True
    assert args.maxdate == "2015-12-31"
    assert args.URL == "https://www.example.org"
    testargs = ["-f", "-min", "2015-12-31"]
    with patch.object(sys, "argv", testargs):
        args = parse_args(testargs)
    assert args.fast is True
    assert args.original is False
    assert args.verbose is False
    assert args.mindate == "2015-12-31"
    # version
    f = io.StringIO()
    testargs = ["", "--version"]
    with pytest.raises(SystemExit) as e, redirect_stdout(f):
        with patch.object(sys, "argv", testargs):
            args = parse_args(testargs)
    assert e.type == SystemExit and e.value.code == 0
    assert re.match(
        r"Htmldate [0-9]\.[0-9]+\.[0-9] - Python [0-9]\.[0-9]+\.[0-9]", f.getvalue()
    )


def test_cli():
    "Test the command-line interface"
    assert examine(" ", extensive_bool=True) is None
    assert examine("0" * int(10e7), extensive_bool=True) is None
    assert examine(" ", False) is None
    assert examine("0" * int(10e7), False) is None
    assert (
        examine(
            '<?xml version="1.0" encoding="utf-8"?><!DOCTYPE html><html lang="en"><head><meta charset="utf-8"/><meta http-equiv="Content-Type" content="text/html; charset=utf-8"/></head><body><p>A</p><p>B</p></body></html>'
        )
        is None
    )
    assert (
        examine(
            '<html><body><span class="entry-date">12. Juli 2016</span></body></html>',
            True,
        )
        == "2016-07-12"
    )
    assert (
        examine("<html><body>2016-07-12</body></html>", extensive_bool=True)
        == "2016-07-12"
    )
    assert (
        examine(
            "<html><body>2016-07-12</body></html>",
            extensive_bool=True,
            maxdate="2015-01-01",
        )
        is None
    )
    assert (
        examine(
            "<html><body>2016-07-12</body></html>",
            extensive_bool=True,
            maxdate="2017-12-31",
        )
        == "2016-07-12"
    )
    assert (
        examine(
            "<html><body>2016-07-12</body></html>",
            extensive_bool=True,
            maxdate="2017-41-41",
        )
        == "2016-07-12"
    )
    # first test
    testargs = ["", "-u", "123", "-v"]
    with patch.object(sys, "argv", testargs):
        args = parse_args(testargs)
    with pytest.raises(SystemExit) as err:
        process_args(args)
    assert err.type == SystemExit
    # meaningful test
    testargs = ["", "-u", "https://httpbin.org/html"]
    with patch.object(sys, "argv", testargs):
        args = parse_args(testargs)
    f = io.StringIO()
    with redirect_stdout(f):
        process_args(args)
    assert len(f.getvalue()) == 0
    # second test
    args.URL = None
    args.inputfile = os.path.join(TEST_DIR, "testlist.txt")
    f = io.StringIO()
    with redirect_stdout(f):
        process_args(args)
    assert f.getvalue() == "https://httpbin.org/html\tNone\n"


def test_download():
    """test page download"""
    with pytest.raises(SystemExit):
        with patch.object(sys, "argv", ["", "-u", "https://httpbin.org/status/404"]):
            main()
    url = "https://httpbin.org/status/200"
    teststring = fetch_url(url)
    assert teststring is None
    assert examine(teststring) is None
    url = "https://httpbin.org/links/2/2"
    teststring = fetch_url(url)
    assert teststring is not None
    assert examine(teststring) is None
    url = "https://httpbin.org/html"
    teststring = fetch_url(url)
    assert teststring is not None
    assert examine(teststring, False) is None


def test_dependencies():
    "Test README examples for consistency"
    if EXT_PARSER is True:
        assert (
            try_date_expr(
                "Fri | September 1 | 2017",
                OUTPUTFORMAT,
                True,
                MIN_DATE,
                LATEST_POSSIBLE,
            )
            == "2017-09-01"
        )


if __name__ == "__main__":

    # function-level
    test_input()
    test_sanity()
    test_date_validator()
    test_search_pattern()
    test_try_date_expr()
    test_convert_date()
    test_compare_reference()
    test_candidate_selection()
    test_regex_parse()
    test_external_date_parser()
    # test_header()

    # module-level
    test_no_date()
    test_exact_date()
    test_search_html()
    test_url()
    test_approximate_url()
    test_idiosyncrasies()
    # new_pages()

    # dependencies
    test_dependencies()

    # cli
    test_parser()
    test_cli()

    # loading functions
    test_download()



# pylint:disable-msg=E0611,I1101
"""
Module bundling functions related to HTML processing.
"""

## This file is available from https://github.com/adbar/htmldate
## under GNU GPL v3 license


# standard
import logging
import re

from typing import Any, List, Optional, Set, Union

import urllib3


# CChardet is faster and can be more accurate
try:
    from cchardet import detect as cchardet_detect  # type: ignore
except ImportError:
    cchardet_detect = None
from charset_normalizer import from_bytes

from lxml.html import HtmlElement, HTMLParser, fromstring  # type: ignore

from .settings import MAX_FILE_SIZE, MIN_FILE_SIZE


LOGGER = logging.getLogger(__name__)

UNICODE_ALIASES: Set[str] = {"utf-8", "utf_8"}

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
RETRY_STRATEGY = urllib3.util.Retry(
    total=3,
    connect=0,
    status_forcelist=[429, 500, 502, 503, 504],
)
HTTP_POOL = urllib3.PoolManager(retries=RETRY_STRATEGY)

HTML_PARSER = HTMLParser(
    collect_ids=False, default_doctype=False, encoding="utf-8", remove_pis=True
)

DOCTYPE_TAG = re.compile("^< ?! ?DOCTYPE.+?/ ?>", re.I)


def isutf8(data: bytes) -> bool:
    """Simple heuristic to determine if a bytestring uses standard unicode encoding"""
    try:
        data.decode("UTF-8")
    except UnicodeDecodeError:
        return False
    else:
        return True


def detect_encoding(bytesobject: bytes) -> List[str]:
    """Read all input or first chunk and return a list of encodings"""
    # alternatives: https://github.com/scrapy/w3lib/blob/master/w3lib/encoding.py
    # unicode-test
    if isutf8(bytesobject):
        return ["utf-8"]
    guesses = []
    # additional module
    if cchardet_detect is not None:
        cchardet_guess = cchardet_detect(bytesobject)["encoding"]
        if cchardet_guess is not None:
            guesses.append(cchardet_guess.lower())
    # try charset_normalizer on first part, fallback on full document
    detection_results = from_bytes(bytesobject[:15000]) or from_bytes(bytesobject)
    # return alternatives
    if len(detection_results) > 0:
        guesses.extend([r.encoding for r in detection_results])
    # it cannot be utf-8 (tested above)
    return [g for g in guesses if g not in UNICODE_ALIASES]


def decode_file(filecontent: Union[bytes, str]) -> str:
    """Guess bytestring encoding and try to decode to Unicode string.
    Resort to destructive conversion otherwise."""
    # init
    if isinstance(filecontent, str):
        return filecontent
    htmltext = None
    # encoding
    for guessed_encoding in detect_encoding(filecontent):
        try:
            htmltext = filecontent.decode(guessed_encoding)
        except (LookupError, UnicodeDecodeError):  # VISCII: lookup
            LOGGER.warning("wrong encoding detected: %s", guessed_encoding)
            htmltext = None
        else:
            break
    # return original content if nothing else succeeded
    return htmltext or str(filecontent, encoding="utf-8", errors="replace")


def decode_response(response: Any) -> str:
    """Read the urllib3 object corresponding to the server response, then
    try to guess its encoding and decode it to return a unicode string"""
    # urllib3 response object / bytes switch
    if isinstance(response, urllib3.response.HTTPResponse) or hasattr(response, "data"):
        resp_content = response.data
    else:
        resp_content = response
    return decode_file(resp_content)


def fetch_url(url: str) -> Optional[str]:
    """Fetches page using urllib3 and decodes the response.

    Args:
        url: URL of the page to fetch.

    Returns:
        HTML code as string, or Urllib3 response object (headers + body), or empty string in case
        the result is invalid, or None if there was a problem with the network.

    """
    # send
    try:
        # read by streaming chunks (stream=True, iter_content=xx)
        # so we can stop downloading as soon as MAX_FILE_SIZE is reached
        response = HTTP_POOL.request("GET", url, timeout=30)  # type: ignore
    except Exception as err:
        LOGGER.error("download error: %s %s", url, err)  # sys.exc_info()[0]
    else:
        # safety checks
        if response.status != 200:
            LOGGER.error("not a 200 response: %s for URL %s", response.status, url)
        elif (
            response.data is None
            or len(response.data) < MIN_FILE_SIZE
            or len(response.data) > MAX_FILE_SIZE
        ):
            LOGGER.error("incorrect input data for URL %s", url)
        else:
            return decode_response(response.data)
    return None


def is_dubious_html(beginning: str) -> bool:
    "Assess if the object is proper HTML (awith a corresponding tag or declaration)."
    return "html" not in beginning


def strip_faulty_doctypes(htmlstring: str, beginning: str) -> str:
    "Repair faulty doctype strings to make then palatable for libxml2."
    # libxml2/LXML issue: https://bugs.launchpad.net/lxml/+bug/1955915
    if "doctype" in beginning:
        firstline, _, rest = htmlstring.partition("\n")
        return DOCTYPE_TAG.sub("", firstline, count=1) + "\n" + rest
    return htmlstring


def fromstring_bytes(htmlobject: str) -> Optional[HtmlElement]:
    "Try to pass bytes to LXML parser."
    tree = None
    try:
        tree = fromstring(htmlobject.encode("utf8"), parser=HTML_PARSER)
    except Exception as err:
        LOGGER.error("lxml parser bytestring %s", err)
    return tree


def load_html(htmlobject: Union[bytes, str, HtmlElement]) -> Optional[HtmlElement]:
    """Load object given as input and validate its type
    (accepted: lxml.html tree, bytestring and string)
    """
    # use tree directly
    if isinstance(htmlobject, HtmlElement):
        return htmlobject
    # do not accept any other type after this point
    if not isinstance(htmlobject, (bytes, str)):
        raise TypeError("incompatible input type: %s", type(htmlobject))
    # the string is a URL, download it
    if isinstance(htmlobject, str) and htmlobject.startswith("http"):
        htmltext = None
        if re.match(r"https?://[^ ]+$", htmlobject):
            LOGGER.info("URL detected, downloading: %s", htmlobject)
            htmltext = fetch_url(htmlobject)
            if htmltext is not None:
                htmlobject = htmltext
        # log the error and quit
        if htmltext is None:
            raise ValueError("URL couldn't be processed: %s", htmlobject)
    # start processing
    tree = None
    # try to guess encoding and decode file: if None then keep original
    htmlobject = decode_file(htmlobject)
    # sanity checks
    beginning = htmlobject[:50].lower()
    check_flag = is_dubious_html(beginning)
    # repair first
    htmlobject = strip_faulty_doctypes(htmlobject, beginning)
    # first pass: use Unicode string
    fallback_parse = False
    try:
        tree = fromstring(htmlobject, parser=HTML_PARSER)
    except ValueError:
        # "Unicode strings with encoding declaration are not supported."
        fallback_parse = True
        tree = fromstring_bytes(htmlobject)
    except Exception as err:  # pragma: no cover
        LOGGER.error("lxml parsing failed: %s", err)
    # second pass: try passing bytes to LXML
    if (tree is None or len(tree) < 2) and fallback_parse is False:
        tree = fromstring_bytes(htmlobject)
    # rejection test: is it (well-formed) HTML at all?
    # log parsing errors
    if tree is not None and check_flag is True and len(tree) < 2:
        LOGGER.error(
            "parsed tree length: %s, wrong data type or not valid HTML", len(tree)
        )
        tree = None
    return tree
